# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ncclient import manager
from ncclient.operations import RPCError, TimeoutExpiredError
from ncclient.xml_ import new_ele, sub_ele, to_ele, to_xml
from netaddr import IPNetwork

from netman import regex
from netman.core.objects.access_groups import IN, OUT
from netman.core.objects.bond import Bond
from netman.core.objects.exceptions import LockedSwitch, VlanAlreadyExist, UnknownVlan, \
    InterfaceInWrongPortMode, UnknownInterface, AccessVlanNotSet, NativeVlanNotSet, TrunkVlanNotSet, VlanAlreadyInTrunk, \
    BadBondNumber, BondAlreadyExist, UnknownBond, InterfaceNotInBond, OperationNotCompleted, InvalidMtuSize
from netman.core.objects.interface import Interface
from netman.core.objects.interface_states import ON, OFF
from netman.core.objects.port_modes import ACCESS, TRUNK, BOND_MEMBER
from netman.core.objects.switch_base import SwitchBase
from netman.core.objects.vlan import Vlan


class Juniper(SwitchBase):

    def __init__(self, switch_descriptor, custom_strategies,
                 timeout=300):
        super(Juniper, self).__init__(switch_descriptor)
        self.timeout = timeout
        self.custom_strategies = custom_strategies
        self.netconf = None

        self.in_transaction = False

    def _connect(self):
        params = dict(
            host=self.switch_descriptor.hostname,
            username=self.switch_descriptor.username,
            password=self.switch_descriptor.password,
            hostkey_verify=False,
            device_params={'name': 'junos'},
            timeout=self.timeout
        )

        if self.switch_descriptor.port:
            params["port"] = self.switch_descriptor.port

        self.netconf = manager.connect(**params)

    def _disconnect(self):
        try:
            self.netconf.close_session()
        except TimeoutExpiredError:
            pass

    def start_transaction(self):
        try:
            self.netconf.lock(target="candidate")
        except RPCError as e:
            if "configuration database modified" in e.message:
                self.rollback_transaction()
                self.netconf.lock(target="candidate")
            elif "Configuration database is already open" in e.message:
                raise LockedSwitch()
            else:
                raise
        self.in_transaction = True

    def end_transaction(self):
        self.in_transaction = False
        self.netconf.unlock(target="candidate")

    def rollback_transaction(self):
        self.netconf.discard_changes()

    def commit_transaction(self):
        try:
            self.netconf.commit()
        except RPCError as e:
            self.logger.info("An RPCError was raised : {}".format(e))
            raise OperationNotCompleted(str(e).strip())

    def get_vlans(self):
        config = self.query(self.custom_strategies.all_vlans, all_interfaces)

        vlan_list = []
        for vlan_node in self.custom_strategies.vlan_nodes(config):
            vlan = self.get_vlan_from_node(vlan_node, config)
            if vlan is not None:
                vlan_list.append(vlan)

        return vlan_list

    def get_vlan(self, number):
        config = self.query(self.custom_strategies.all_vlans, all_interfaces)
        vlan_node = self.custom_strategies.vlan_node(config, number)
        return self.get_vlan_from_node(vlan_node, config)

    def get_vlan_from_node(self, vlan_node, config):
        vlan_id_node = first(vlan_node.xpath("vlan-id"))

        vlan = None
        if vlan_id_node is not None:
            vlan = Vlan(number=int(vlan_id_node.text), icmp_redirects=True)

            description_node = first(vlan_node.xpath("description"))
            if description_node is not None:
                vlan.name = description_node.text

            l3_if_type, l3_if_name = self.custom_strategies.get_l3_interface(vlan_node)
            if l3_if_name is not None:
                interface_vlan_node = first(config.xpath("data/configuration/interfaces/interface/name[text()=\"{}\"]/.."
                                                         "/unit/name[text()=\"{}\"]/..".format(l3_if_type, l3_if_name)))
                if interface_vlan_node is not None:
                    vlan.ips = parse_ips(interface_vlan_node)
                    vlan.access_groups[IN] = parse_inet_filter(interface_vlan_node, "input")
                    vlan.access_groups[OUT] = parse_inet_filter(interface_vlan_node, "output")
                    vlan.vrrp_groups = self.custom_strategies.parse_vrrp_groups(interface_vlan_node)
                    vlan.icmp_redirects = self.custom_strategies.parse_icmp_redirects(interface_vlan_node)
        return vlan

    def get_interfaces(self):
        physical_interfaces = self._list_physical_interfaces()
        config = self.query(all_interfaces, self.custom_strategies.all_vlans)

        interface_list = []
        for phys_int in physical_interfaces:
            if not phys_int.name.startswith("ae"):
                interface_node = first(config.xpath("data/configuration/interfaces/interface/name[text()=\"{}\"]/.."
                                                    .format(phys_int.name)))
                if interface_node is not None:
                    interface_list.append(self.node_to_interface(interface_node, config))
                else:
                    interface_list.append(phys_int.to_interface())

        return interface_list

    def add_vlan(self, number, name=None):
        config = self.query(self.custom_strategies.all_vlans)

        try:
            self.custom_strategies.vlan_node(config, number)
            raise VlanAlreadyExist(number)
        except UnknownVlan:
            pass

        update = Update()
        self.custom_strategies.add_update_vlans(update, number, name)

        try:
            self._push(update)
        except RPCError as e:
            self.custom_strategies.manage_update_vlan_exception(e.message, number)
            raise

    def remove_vlan(self, number):
        config = self.query(self.custom_strategies.all_vlans, all_interfaces)

        vlan_node = self.custom_strategies.vlan_node(config, number)
        vlan_name = first(vlan_node.xpath("name")).text

        update = Update()
        self.custom_strategies.remove_update_vlans(update, vlan_name)

        l3_if_type, l3_if_name = self.custom_strategies.get_l3_interface(vlan_node)
        if l3_if_name is not None:
            update.add_interface(interface_unit_interface_removal(l3_if_type, l3_if_name))

        for interface_node in config.xpath("data/configuration/interfaces/interface"):
            members_modifications = self.custom_strategies.craft_members_modification_to_remove_vlan(interface_node, vlan_name, number)

            if len(members_modifications) > 0:
                update.add_interface(self.custom_strategies.interface_vlan_members_update(
                    first(interface_node.xpath("name")).text,
                    first(interface_node.xpath("unit/name")).text,
                    members_modifications)
                )

        self._push(update)

    def set_access_mode(self, interface_id):
        update_attributes = []

        config = self.query(all_interfaces, self.custom_strategies.all_vlans)

        interface_node = self.get_interface_config(interface_id, config)

        interface = self.node_to_interface(interface_node, config)

        if self.get_port_mode(interface_node) in (TRUNK, None):
            update_attributes.append(self.custom_strategies.get_interface_port_mode_update_element("access"))

        if len(interface.trunk_vlans) > 0:
            update_attributes.append(self.custom_strategies.get_delete_trunk_vlan_element())

        if interface.trunk_native_vlan is not None:
            update_attributes.append(to_ele('<native-vlan-id operation="delete" />'))

        if len(update_attributes) > 0:
            update = Update()
            update.add_interface(self.custom_strategies.interface_update(interface_id, "0", update_attributes))
            self._push_interface_update(interface_id, update)

    def set_trunk_mode(self, interface_id):
        update_attributes = []

        config = self.query(one_interface(interface_id), self.custom_strategies.all_vlans)
        interface_node = self.get_interface_config(interface_id, config)
        interface = self.node_to_interface(interface_node, config)

        if interface.port_mode is ACCESS or interface.port_mode is None:
            update_attributes.append(self.custom_strategies.get_interface_port_mode_update_element("trunk"))

        if interface.access_vlan is not None:
            update_attributes.append(self.custom_strategies.get_delete_vlan_element())

        if len(update_attributes) > 0:
            update = Update()
            update.add_interface(self.custom_strategies.interface_update(interface_id, "0", update_attributes))

            self._push_interface_update(interface_id, update)

    def set_access_vlan(self, interface_id, vlan):
        update_attributes = []
        update_vlan_members = []

        config = self.query(all_interfaces, self.custom_strategies.all_vlans)

        self.custom_strategies.vlan_node(config, vlan)

        interface_node = self.get_interface_config(interface_id, config)
        interface = self.node_to_interface(interface_node, config)

        if interface.port_mode == TRUNK:
            raise InterfaceInWrongPortMode("trunk")
        elif self.get_port_mode(interface_node) is None:
            update_attributes.append(self.custom_strategies.get_interface_port_mode_update_element("access"))

        if interface.access_vlan != vlan:
            self.custom_strategies.update_vlan_members(interface_node, update_vlan_members, vlan)

        if update_attributes or update_vlan_members:
            update = Update()
            update.add_interface(self.custom_strategies.interface_update(interface_id, "0", update_attributes, update_vlan_members))

            try:
                self._push_interface_update(interface_id, update)
            except RPCError as e:
                if "No vlan matches vlan tag" in e.message:
                    raise UnknownVlan(vlan)
                raise

    def unset_interface_access_vlan(self, interface_id):
        config = self.query(one_interface(interface_id), self.custom_strategies.all_vlans)
        interface_node = self.get_interface_config(interface_id, config)
        interface = self.node_to_interface(interface_node, config)

        if interface.port_mode == TRUNK:
            raise InterfaceInWrongPortMode("trunk")

        if interface.access_vlan is not None:
            update = Update()
            update.add_interface(self.custom_strategies.interface_update(interface_id, "0", [to_ele('<vlan operation="delete" />')]))

            self._push(update)
        else:
            raise AccessVlanNotSet(interface_id)

    def set_interface_native_vlan(self, interface_id, vlan):
        port_mode_node = None
        native_vlan_id_node = None

        config = self.query(all_interfaces, self.custom_strategies.all_vlans)

        self.custom_strategies.vlan_node(config, vlan)

        interface_node = self.get_interface_config(interface_id, config)

        interface = self.node_to_interface(interface_node, config)

        actual_port_mode = self.get_port_mode(interface_node)
        if actual_port_mode is ACCESS:
            raise InterfaceInWrongPortMode("access")
        elif actual_port_mode is None:
            port_mode_node = self.custom_strategies.get_interface_port_mode_update_element("trunk")

        if vlan in interface.trunk_vlans:
            raise VlanAlreadyInTrunk(vlan)
        elif interface.trunk_native_vlan != vlan:
            native_vlan_id_node = to_ele("<native-vlan-id>{}</native-vlan-id>".format(vlan))

        if native_vlan_id_node is not None:

            interface = self.custom_strategies.interface_update(interface_id, "0", [port_mode_node] if port_mode_node is not None else [])
            self.custom_strategies.set_native_vlan_id_node(interface, native_vlan_id_node)
            update = Update()
            update.add_interface(interface)

            try:
                self._push_interface_update(interface_id, update)
            except RPCError as e:
                if "No vlan matches vlan tag" in e.message:
                    raise UnknownVlan(vlan)
                raise

    def set_interface_auto_negotiation_state(self, interface_id, negotiation_state):
        content = to_ele("""
            <interface>
                <name>{0}</name>
            </interface>
            """.format(interface_id))
        if negotiation_state == ON:
            content.append(to_ele("<ether-options><auto-negotiation/></ether-options>"))
        else:
            content.append(to_ele("<ether-options><no-auto-negotiation/></ether-options>"))
        update = Update()
        update.add_interface(content)

        self._push_interface_update(interface_id, update)

    def unset_interface_auto_negotiation_state(self, interface_id):
        config = self.query(one_interface(interface_id))
        interface_node = self.get_interface_config(interface_id, config)

        if interface_node is None:
            self._get_physical_interface(interface_id)
            return

        auto_negotiation_present = first(interface_node.xpath('ether-options/auto-negotiation')) is not None
        no_auto_negotiation_present = first(interface_node.xpath('ether-options/no-auto-negotiation')) is not None

        if auto_negotiation_present or no_auto_negotiation_present:
            content = to_ele("""
                <interface>
                    <name>{0}</name>
                </interface>
            """.format(interface_id))
            ether_options = to_ele("<ether-options/>")
            if auto_negotiation_present:
                ether_options.append(to_ele("<auto-negotiation operation=\"delete\"/>"))
            elif no_auto_negotiation_present:
                ether_options.append(to_ele("<no-auto-negotiation operation=\"delete\"/>"))
            update = Update()

            content.append(ether_options)
            update.add_interface(content)

            self._push_interface_update(interface_id, update)

    def reset_interface(self, interface_id):
        content = to_ele("""
            <interface operation=\"delete\">
                <name>{0}</name>
            </interface>
        """.format(interface_id))
        update = Update()
        update.add_interface(content)

        self._push_interface_update(interface_id, update)

    def unset_interface_native_vlan(self, interface_id):
        config = self.query(one_interface(interface_id), self.custom_strategies.all_vlans)
        interface_node = self.get_interface_config(interface_id, config)
        interface = self.node_to_interface(interface_node, config)

        if interface.trunk_native_vlan is None:
            raise NativeVlanNotSet(interface_id)

        update = Update()
        update.add_interface(self.custom_strategies.interface_update(interface_id, "0", [to_ele("<native-vlan-id operation=\"delete\" />")]))

        self._push(update)

    def add_trunk_vlan(self, interface_id, vlan):
        config = self.query(all_interfaces, self.custom_strategies.all_vlans)

        self.custom_strategies.vlan_node(config, vlan)

        interface_node = self.get_interface_config(interface_id, config)

        interface = self.node_to_interface(interface_node, config)

        actual_port_mode = self.get_port_mode(interface_node)

        if actual_port_mode is ACCESS or interface.access_vlan is not None:
            raise InterfaceInWrongPortMode("access")

        if vlan not in interface.trunk_vlans:
            update = Update()
            update.add_interface(self.custom_strategies.interface_update(
                interface_id, "0",
                [self.custom_strategies.get_interface_port_mode_update_element("trunk")] if actual_port_mode is None else None,
                [self.custom_strategies.get_vlan_member_update_element(vlan)]
            ))

            self._push_interface_update(interface_id, update)

    def remove_trunk_vlan(self, interface_id, vlan):
        config = self.query(all_interfaces, self.custom_strategies.all_vlans)
        interface_node = self.get_interface_config(interface_id, config)
        if interface_node is None:
            raise UnknownInterface(interface_id)

        interface = self.node_to_interface(interface_node, config)

        if interface.port_mode is ACCESS:
            raise InterfaceInWrongPortMode("access")

        vlan_node = self.custom_strategies.vlan_node(config, vlan)
        vlan_name = first(vlan_node.xpath("name")).text

        modifications = self.custom_strategies.craft_members_modification_to_remove_vlan(interface_node, vlan_name, vlan)
        if len(modifications) == 0:
            raise TrunkVlanNotSet(interface_id)

        update = Update()
        update.add_interface(self.custom_strategies.interface_update(interface_id, "0", vlan_members=modifications))

        self._push(update)

    def set_interface_description(self, interface_id, description):
        update = Update()
        update.add_interface(interface_main_update(interface_id, [
            to_ele("<description>{}</description>".format(description))
        ]))

        try:
            self._push(update)
        except RPCError as e:
            self.logger.info("actual setting error was {}".format(e))
            raise UnknownInterface(interface_id)

    def unset_interface_description(self, interface_id):
        update = Update()
        update.add_interface(interface_main_update(interface_id, [
            to_ele("<description operation=\"delete\" />")
        ]))

        try:
            self._push(update)
        except RPCError as e:
            if e.severity != "warning":
                raise UnknownInterface(interface_id)

    def set_interface_mtu(self, interface_id, size):
        update = Update()
        update.add_interface(interface_main_update(interface_id, [
            to_ele("<mtu>{}</mtu>".format(size))
        ]))

        try:
            self._push(update)
        except RPCError as e:
            self.logger.info("actual setting error was {}".format(e))
            if "Value {} is not within range".format(size) in str(e):
                raise InvalidMtuSize(str(e))

            raise UnknownInterface(interface_id)

    def unset_interface_mtu(self, interface_id):
        update = Update()
        update.add_interface(interface_main_update(interface_id, [
            to_ele("<mtu operation=\"delete\" />")
        ]))

        try:
            self._push(update)
        except RPCError as e:
            if e.severity != "warning":
                raise UnknownInterface(interface_id)

    def edit_interface_spanning_tree(self, interface_id, edge=None):
        config = self.query(one_interface(interface_id),
                            one_protocol_interface("rstp", self._for_protocol(interface_id)))

        if edge is not None:
            modifications = _compute_edge_state_modifications(interface_id, edge, config)

            if modifications:
                update = Update()
                update.add_protocol_interface("rstp", to_ele("""
                   <interface>
                     <name>{}</name>
                     {}
                   </interface>
                """.format(interface_id, "".join(modifications))))

                self._push_interface_update(interface_id, update)

    def set_interface_state(self, interface_id, state):
        update = Update()
        update.add_interface(interface_state_update(interface_id, state))

        try:
            self._push(update)
        except RPCError as e:
            self.logger.info("actual setting error was {}".format(e))
            # When sending a "delete operation" on a nonexistent element <disable />, this is the error that is thrown.
            # It's ignored because the result of this operation would be the same as if the command was successful.
            if "statement not found" not in e.message:
                raise UnknownInterface(interface_id)

    def unset_interface_state(self, interface_id):
        self.set_interface_state(interface_id, state=ON)

    def set_interface_lldp_state(self, interface_id, enabled):
        config = self.query(one_interface(interface_id),
                            one_protocol_interface("lldp", self._for_protocol(interface_id)))
        self.get_interface_config(interface_id, config)

        update_ele = None
        disabled_node = first(config.xpath("data/configuration/protocols/lldp/interface/name"
                                           "[text()=\"{0:s}\"]/../disable".format(self._for_protocol(interface_id))))
        if enabled:
            update_ele = protocol_interface_update(interface_id)
            if disabled_node is not None:
                update_ele.append(to_ele('<disable operation="delete"/>'))
        elif not enabled and disabled_node is None:
            update_ele = protocol_interface_update(interface_id)
            update_ele.append(to_ele('<disable/>'))

        if update_ele is not None:
            update = Update()
            update.add_protocol_interface("lldp", update_ele)

            self._push_interface_update(interface_id, update)

    def add_bond(self, number):
        config = self.query(one_interface(bond_name(number)))
        if len(config.xpath("data/configuration/interfaces/interface")) > 0:
            raise BondAlreadyExist(number)

        update = Update()
        update.add_interface(bond_update(number, bond_lacp_options()))

        try:
            self._push(update)
        except RPCError as e:
            if "device value outside range" in e.message:
                raise BadBondNumber()

            raise

    def remove_bond(self, number):
        config = self.query(all_interfaces, one_protocol_interface("rstp", self._for_protocol(bond_name(number))))
        self.get_bond_config(number, config)

        update = Update()
        update.add_interface(interface_removal(bond_name(number)))

        rstp_node = first(config.xpath("data/configuration/protocols/rstp/interface/name[text()=\"{0:s}\"]/.."
                                       .format(self._for_protocol(bond_name(number)))))
        if rstp_node is not None:
            update.add_protocol_interface("rstp", rstp_interface_removal(bond_name(number)))

        for interface_node in self.get_bond_slaves_config(number, config):
            interface_name = first(interface_node.xpath("name")).text
            update.add_interface(free_from_bond_operation(interface_name))

        self._push(update)

    def add_interface_to_bond(self, interface, bond_id):
        config = self.query(all_interfaces, self.custom_strategies.all_vlans, rstp_protocol_interfaces)
        bond = self.node_to_bond(self.get_bond_config(bond_id, config), config)

        update = Update()
        self.custom_strategies.add_enslave_to_bond_operations(update, interface, bond)

        rstp_node = first(config.xpath("data/configuration/protocols/rstp/interface/name[text()=\"{0:s}\"]"
                                       .format(self._for_protocol(interface))))
        if rstp_node is not None:
            update.add_protocol_interface("rstp", rstp_interface_removal(interface))

        self._push_interface_update(interface, update)

    def remove_interface_from_bond(self, interface):
        try:
            update = Update()
            update.add_interface(free_from_bond_operation(interface))
            self._push(update)
        except RPCError:
            self._get_physical_interface(interface)

            raise InterfaceNotInBond()

    def set_bond_link_speed(self, number, speed):
        config = self.query(all_interfaces)
        self.get_bond_config(number, config)

        update = Update()
        update.add_interface(bond_update(number, bond_link_speed(speed)))

        self.custom_strategies.add_update_bond_members_speed_operations(
            update,
            self.get_bond_slaves_config(number, config),
            speed
        )

        self._push(update)

    def get_bond(self, number):
        config = self.query(all_interfaces, self.custom_strategies.all_vlans)

        bond_node = self.get_bond_config(number, config)
        return self.node_to_bond(bond_node, config, self.get_bond_slaves_config(
            value_of(bond_node.xpath("name"), transformer=bond_number), config))

    def get_bonds(self):
        config = self.query(all_interfaces, self.custom_strategies.all_vlans)
        bond_nodes = config.xpath("data/configuration/interfaces/interface/aggregated-ether-options/..")
        return [
            self.node_to_bond(node, config, self.get_bond_slaves_config(
                value_of(node.xpath("name"), transformer=bond_number), config))
            for node in bond_nodes]

    def set_bond_description(self, number, description):
        return self.set_interface_description(bond_name(number), description)

    def unset_bond_description(self, number):
        return self.unset_interface_description(bond_name(number))

    def set_bond_mtu(self, number, size):
        return self.set_interface_mtu(bond_name(number), size)

    def unset_bond_mtu(self, number):
        return self.unset_interface_mtu(bond_name(number))

    def set_bond_trunk_mode(self, number):
        return self.set_trunk_mode(bond_name(number))

    def set_bond_access_mode(self, number):
        return self.set_access_mode(bond_name(number))

    def add_bond_trunk_vlan(self, number, vlan):
        return self.add_trunk_vlan(bond_name(number), vlan)

    def remove_bond_trunk_vlan(self, number, vlan):
        return self.remove_trunk_vlan(bond_name(number), vlan)

    def set_bond_native_vlan(self, number, vlan):
        return self.set_interface_native_vlan(bond_name(number), vlan)

    def unset_bond_native_vlan(self, number):
        return self.unset_interface_native_vlan(bond_name(number))

    def edit_bond_spanning_tree(self, number, edge=None):
        return self.edit_interface_spanning_tree(bond_name(number), edge=edge)

    def _push_interface_update(self, interface_id, configuration):
        try:
            self._push(configuration)
        except RPCError as e:
            if "port value outside range" in e.message \
                    or "invalid interface type" in e.message \
                    or "device value outside range" in e.message:
                raise UnknownInterface(interface_id)
            raise

    def _push(self, configuration):
        config = new_ele('config')
        config.append(configuration.root)

        self.logger.info("Sending edit : {}".format(to_xml(config)))
        try:
            self.netconf.edit_config(target="candidate", config=config)
        except RPCError as e:
            self.logger.info("An RPCError was raised : {}".format(e))
            raise

    def query(self, *args):
        filter_node = new_ele("filter")
        conf = sub_ele(filter_node, "configuration")
        for arg in args:
            conf.append(arg())
        return self.netconf.get_config(source="candidate" if self.in_transaction else "running", filter=filter_node)

    def get_interface(self, interface_id):
        config = self.query(one_interface(interface_id), self.custom_strategies.all_vlans)
        interface_node = self.get_interface_config(interface_id, config)
        if interface_node is not None:
            return self.node_to_interface(interface_node, config)

        return self._get_physical_interface(interface_id).to_interface()

    def get_interface_config(self, interface_id, config=None):
        config = config or self.query(one_interface(interface_id))
        interface_node = first(config.xpath(
            "data/configuration/interfaces/interface/name[text()=\"{0:s}\"]/..".format(interface_id)))
        return interface_node

    def get_bond_config(self, number, config):
        interface_node = first(config.xpath("data/configuration/interfaces/interface/name[text()=\"{}\"]/..".format(bond_name(number))))
        if interface_node is None:
            raise UnknownBond(number)
        return interface_node

    def get_bond_slaves_config(self, bond_id, config=None):
        config = config or self.query(all_interfaces)

        return config.xpath(
            'data/configuration/interfaces/interface/ether-options/'
            'ieee-802.3ad/bundle[text()=\"{0}\"]/../../..'.format(
                bond_name(bond_id)))

    def get_port_mode(self, interface_node):
        if interface_node is None:
            return None

        if get_bond_master(interface_node) is not None:
            return BOND_MEMBER
        actual_port_mode_node = first(self.custom_strategies.get_port_mode_node_in_inteface_node(interface_node))
        if actual_port_mode_node is None:
            return None
        else:
            return {"access": ACCESS, "trunk": TRUNK}[actual_port_mode_node.text]

    def fill_interface_from_node(self, interface, interface_node, config):
        if interface_node is not None:
            interface.port_mode = self.get_port_mode(interface_node) or ACCESS
            vlans = self.custom_strategies.list_vlan_members(interface_node, config)
            if interface.port_mode is ACCESS:
                interface.access_vlan = first(vlans)
            else:
                interface.trunk_vlans = vlans
            interface.trunk_native_vlan = value_of(self.custom_strategies.get_interface_trunk_native_vlan_id_node(interface_node), transformer=int)
            interface.shutdown = first(interface_node.xpath("disable")) is not None
            interface.mtu = value_of(interface_node.xpath("mtu"), transformer=int)
            if first(interface_node.xpath('ether-options/auto-negotiation')) is not None:
                interface.auto_negotiation = True
            elif first(interface_node.xpath('ether-options/no-auto-negotiation')) is not None:
                interface.auto_negotiation = False
        return interface

    def node_to_interface(self, interface_node, config):
        interface = Interface()
        if interface_node is not None:
            interface.name = value_of(interface_node.xpath("name"))
            interface.bond_master = get_bond_master(interface_node)
        self.fill_interface_from_node(interface, interface_node, config)
        return interface

    def node_to_bond(self, bond_node, config, member_nodes=None):
        member_nodes = member_nodes or []
        bond = Bond(
            number=value_of(bond_node.xpath("name"), transformer=bond_number),
            link_speed=first_text(bond_node.xpath("aggregated-ether-options/link-speed")),
            members=[first_text(member.xpath('name')) for member in member_nodes]
        )
        self.fill_interface_from_node(bond, bond_node, config)
        return bond

    def get_vlan_interfaces(self, vlan_number):
        config = self.query(self.custom_strategies.one_vlan_by_vlan_id(vlan_number), all_interfaces)

        vlan_node = self.custom_strategies.vlan_node(config, vlan_number)
        interface_nodes = config.xpath("data/configuration/interfaces/interface")
        return self.get_vlan_interfaces_from_node(vlan_node, interface_nodes)

    def get_vlan_interfaces_from_node(self, vlan_node, interface_nodes):
        vlan_name = first(vlan_node.xpath("name")).text
        vlan_number = int(first(vlan_node.xpath("vlan-id")).text)
        interfaces = []
        for interface in interface_nodes:
            native_vlan_id_node = self.custom_strategies.get_interface_trunk_native_vlan_id_node(interface)
            if len(native_vlan_id_node) == 1 and int(first(native_vlan_id_node).text) == vlan_number:
                interfaces.append(first(interface.xpath("name")).text)

            members = [members.text for members in interface.xpath("unit/family/ethernet-switching/vlan/members")]
            if _is_vlan_in_interface_members(vlan_number, vlan_name, members):
                interfaces.append(first(interface.xpath("name")).text)
        return interfaces

    def _get_physical_interface(self, interface_id):
        try:
            return next(i for i in self._list_physical_interfaces() if i.name == interface_id)
        except StopIteration:
            raise UnknownInterface(interface_id)

    def _list_physical_interfaces(self):
        terse = self.netconf.rpc(to_ele("""
            <get-interface-information>
              <terse/>
            </get-interface-information>
        """))

        return [_PhysicalInterface(i.xpath("name")[0].text.strip(),
                                   shutdown=i.xpath("admin-status")[0].text.strip() == "down")
                for i in terse.xpath("interface-information/physical-interface")]

    def _for_protocol(self, interface_name):
        return self.custom_strategies.get_protocols_interface_name(interface_name)


def all_interfaces():
    return new_ele("interfaces")


def one_interface(interface_id):
    def m():
        return to_ele("""
            <interfaces>
                <interface>
                    <name>{}</name>
                </interface>
            </interfaces>
        """.format(interface_id))

    return m


def rstp_protocol_interfaces():
    return to_ele("""
        <protocols>
          <rstp>
            <interface />
          </rstp>
        </protocols>
    """)


def one_protocol_interface(protocol, interface_name):
    def m():
        return to_ele("""
            <protocols>
              <{protocol}>
                <interface>
                    <name>{interface}</name>
                </interface>
              </{protocol}>
            </protocols>
        """.format(interface=interface_name, protocol=protocol))

    return m


class Update(object):
    def __init__(self):
        self.root = new_ele("configuration")
        self.vlans_root = None
        self.interfaces_root = None
        self.protocols_root = None
        self.sub_protocol_roots = {}

    def add_vlan(self, vlan, root_ele):
        if self.vlans_root is None:
            self.vlans_root = sub_ele(self.root, root_ele)
        self.vlans_root.append(vlan)

    def add_interface(self, interface):
        if self.interfaces_root is None:
            self.interfaces_root = sub_ele(self.root, "interfaces")
        self.interfaces_root.append(interface)

    def add_protocol(self, protocol):
        if self.protocols_root is None:
            self.protocols_root = sub_ele(self.root, "protocols")
        self.protocols_root.append(protocol)

    def add_protocol_interface(self, protocol, interface):
        if protocol not in self.sub_protocol_roots:
            self.sub_protocol_roots[protocol] = to_ele("<{0}></{0}>".format(protocol))
            self.add_protocol(self.sub_protocol_roots[protocol])
        self.sub_protocol_roots[protocol].append(interface)


def bond_update(number, *aggregated_ether_options):
    content = to_ele("""
        <interface>
            <name>{0}</name>
            <aggregated-ether-options>
            </aggregated-ether-options>
        </interface>
    """.format(bond_name(number)))

    aggregated_ether_options_node = first(content.xpath("//aggregated-ether-options"))
    map(aggregated_ether_options_node.append, aggregated_ether_options)

    return content


def bond_lacp_options():
    return to_ele("""
        <lacp>
            <active/>
            <periodic>slow</periodic>
        </lacp>
    """)


def bond_link_speed(link_speed):
    return to_ele("""
        <link-speed>{0}</link-speed>
    """.format(link_speed))


def vlan_update(number, description):
    content = to_ele("""
        <vlan>
            <name>VLAN{0}</name>
            <vlan-id>{0}</vlan-id>
        </vlan>
    """.format(number))

    if description is not None:
        content.append(to_ele("<description>{}</description>".format(description)))
    return content


def interface_removal(name):
    return to_ele("""
        <interface operation="delete">
            <name>{}</name>
        </interface>""".format(name))


def interface_state_update(name, state):
    interface_state = """<disable />""" if state is OFF else """<disable operation="delete" />"""
    return to_ele("""
        <interface>
            <name>{}</name>
            {}
        </interface>""".format(name, interface_state))


def interface_unit_interface_removal(interface, unit):
    return to_ele("""
        <interface>
          <name>{}</name>
          <unit operation="delete">
            <name>{}</name>
          </unit>
        </interface>
        """.format(interface, unit))


def rstp_interface_removal(interface_id):
    return to_ele("""
        <interface operation="delete" >
            <name>{}</name>
        </interface>
        """.format(interface_id))


def interface_main_update(name, attributes):
    content = to_ele("""
        <interface>
            <name>{}</name>
        </interface>
        """.format(name))

    for attribute in (attributes if attributes is not None else []):
        content.append(attribute)

    return content


def first(node):
    return node[0] if node else None


def first_text(node):
    return node[0].text if node else None


def parse_range(r):
    if regex.match("(\d+)-(\d+)", r):
        return range(int(regex[0]), int(regex[1]) + 1)
    elif regex.match("(\d+)", r):
        return [int(regex[0])]
    return []


def to_range(number_list):
    if len(number_list) > 1:
        return "{}-{}".format(number_list[0], number_list[-1])
    else:
        return str(number_list[0])


def parse_ips(interface_unit_node):
    return sorted(
        [IPNetwork(ip_node.text) for ip_node in interface_unit_node.xpath("family/inet/address/name")],
        key=lambda ip: (ip.value, ip.prefixlen)
    )


def parse_inet_filter(interface_unit_node, direction):
    val = None
    ac_in_node = first(interface_unit_node.xpath("family/inet/filter/{}/filter-name".format(direction)))
    if ac_in_node is not None:
        val = ac_in_node.text
    return val


def bond_name(number):
    return "ae{0}".format(number)


def bond_number(name):
    return int(name[2:])


def free_from_bond_operation(interface_name):
    return to_ele("""
        <interface>
            <name>{0}</name>
            <ether-options>
                <ieee-802.3ad operation=\"delete\" />
            </ether-options>
        </interface>
        """.format(interface_name))


def interface_replace(interface_name, *ether_options):
    content = to_ele("""
        <interface operation=\"replace\">
            <name>{0}</name>
            <ether-options>
            </ether-options>
        </interface>
    """.format(interface_name))

    ether_options_node = first(content.xpath("//ether-options"))
    map(ether_options_node.append, ether_options)

    return content


def interface_speed(speed):
    return to_ele("""
        <speed>
            <ethernet-{0}/>
        </speed>
    """.format(speed))


def interface_speed_update(interface_name, speed):
    return to_ele("""
        <interface>
            <name>{0}</name>
            <ether-options>
                <speed>
                    <ethernet-{1}/>
                </speed>
            </ether-options>
        </interface>
    """.format(interface_name, speed))


def protocol_interface_update(name):
    return to_ele("""
        <interface>
          <name>{}</name>
        </interface>
    """.format(name))


def get_bond_master(interface_node):
    return value_of(
        interface_node.xpath('ether-options/ieee-802.3ad/bundle'),
        transformer=bond_number)


def value_of(xpath_result, transformer=None):
    node = first(xpath_result)
    if node is not None:
        return node.text if transformer is None else transformer(node.text)
    else:
        return None


def _compute_edge_state_modifications(interface_id, edge, config):
    modifications = []
    rstp_node = first(config.xpath("data/configuration/protocols/rstp/interface/name[text()=\"{0:s}.0\"]/.."
                                   .format(interface_id)))

    edge_node = None
    no_root_port_node = None
    if rstp_node is not None:
        edge_node = first(rstp_node.xpath("edge"))
        no_root_port_node = first(rstp_node.xpath("no-root-port"))

    if edge is True:
        if edge_node is None:
            modifications.append("<edge />")
        if no_root_port_node is None:
            modifications.append("<no-root-port />")
    elif edge is False:
        if edge_node is not None:
            modifications.append("<edge operation=\"delete\" />")
        if no_root_port_node is not None:
            modifications.append("<no-root-port operation=\"delete\" />")

    return modifications


def _is_vlan_in_interface_members(vlan_number, vlan_name, members):
    for member_name in members:
        if regex.match("(\d+)-(\d+)", member_name):
            start, end = regex
            if int(start) <= vlan_number <= int(end):
                return True
        elif member_name == vlan_name:
            return True
        elif int(member_name) == vlan_number:
            return True

    return False


class _PhysicalInterface(object):
    def __init__(self, name, shutdown):
        self.name = name
        self.shutdown = shutdown

    def to_interface(self):
        return Interface(name=self.name, shutdown=self.shutdown, port_mode=ACCESS)



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from hamcrest import assert_that, is_, contains_string

from netman.core.objects.exceptions import UnknownInterface, InvalidMtuSize
from tests import has_message
from tests.adapters.compliance_test_case import ComplianceTestCase


class SetInterfaceMtuTest(ComplianceTestCase):
    _dev_sample = "juniper"

    def test_sets_the_mtu_for_the_interface(self):
        self.client.set_interface_mtu(self.test_port, 5000)
        interface = self.client.get_interface(self.test_port)

        assert_that(interface.mtu, is_(5000))

    def test_fails_if_the_value_is_not_within_the_supported_range(self):
        with self.assertRaises(InvalidMtuSize) as expect:
            self.client.set_interface_mtu(self.test_port, 1)

        assert_that(str(expect.exception), contains_string("MTU value is invalid"))

    def test_fails_if_the_interface_does_not_exist(self):
        with self.assertRaises(UnknownInterface) as expect:
            self.client.set_interface_mtu('ge-0/0/128', 5000)

        assert_that(expect.exception, has_message("Unknown interface ge-0/0/128"))

    def tearDown(self):
        self.janitor.unset_interface_mtu(self.test_port)
        super(SetInterfaceMtuTest, self).tearDown()



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from fake_switches.switch_configuration import AggregatedPort
from hamcrest import assert_that, is_
from netaddr import IPNetwork

from tests.adapters.compliance_test_case import ComplianceTestCase


class GetInterfacesTest(ComplianceTestCase):
    _dev_sample = "juniper"

    def test_lists_all_physical_interfaces(self):

        interfaces = self.client.get_interfaces()

        assert_that([i.name for i in interfaces], is_(self._all_physical_test_ports()))

    def test_does_not_list_bonds(self):
        self.try_to.add_bond(1)

        interfaces = self.client.get_interfaces()

        assert_that([i.name for i in interfaces], is_(self._all_physical_test_ports()))

    def _all_physical_test_ports(self):
        return [p.name for p in self.test_ports if not isinstance(p, AggregatedPort)]

    def test_does_not_list_interface_vlans(self):
        self.client.add_vlan(1000)
        self.try_to.add_ip_to_vlan(1000, IPNetwork("1.1.1.1/27"))

        interfaces = self.client.get_interfaces()

        assert_that([i.name for i in interfaces], is_(self._all_physical_test_ports()))

    def tearDown(self):
        self.janitor.remove_bond(1)
        self.janitor.remove_vlan(1000)
        super(GetInterfacesTest, self).tearDown()



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License

import MockSSH

from fake_switches.terminal.telnet import StatefulTelnet


class MockTelnet(StatefulTelnet):
    """
    This aims to act as a telent server over the SAME contract as MockSSH
    Taking the same commands and making them work the same ensure
    the same test code can be applied to SSH and Telnet
    """

    def __init__(self, prompt, commands):
        self._password_input = False

        super(MockTelnet, self).__init__()
        self.prompt = prompt
        self.commands = {c.name: c for c in commands}
        self.cmdstack = []
        self.password_input = False
        self.terminal = self

        self.commands["_exit"] = ExitCommand("_exit")

    @property
    def password_input(self):
        return self._password_input

    @password_input.setter
    def password_input(self, value):
        if self._password_input != value:
            self._password_input = value
            if self._password_input:
                self.enable_input_replacement("")
            else:
                self.disable_input_replacement()

    def connectionMade(self):
        super(MockTelnet, self).connectionMade()
        self.keyHandlers = self._key_handlers
        self.write('Username:')
        self.handler = self.validate_username

    def loseConnection(self):
        self.connectionLost("requested")

    def validate_username(self, _):
        self.write('Password:')
        self.handler = self.validate_password

    def validate_password(self, _):
        self.cmdstack = [RootCommand(self)]
        self.cmdstack[0].resume()
        self.handler = self.command

    def show_prompt(self):
        self.write(self.prompt)

    def command(self, data):
        line = data.rstrip()
        if len(self.cmdstack) > 1:
            self.cmdstack[-1].lineReceived(line)
        else:
            parts = line.split(" ")
            command = parts[0]
            if command in self.commands:
                self.call_command(self.commands[command], parts)

    def call_command(self, cmd, parts=None):
        cmd(self)
        cmd.args = parts
        self.cmdstack.append(cmd)
        cmd.start()

    def nextLine(self):
        return self.next_line()


class RootCommand(object):
    def __init__(self, terminal):
        self.terminal = terminal

    def resume(self):
        self.terminal.show_prompt()


class ExitCommand(MockSSH.command_exit):
    def __init__(self, name, *args):
        self.name = name

    def __call__(self, protocol, *args):
        MockSSH.SSHCommand.__init__(self, protocol, self.name, *args)
        return self



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netman.core.objects import Model
from netman.core.objects.access_groups import OUT, IN


class Vlan(Model):
    def __init__(self, number=None, name=None, ips=None, vrrp_groups=None, vrf_forwarding=None, access_group_in=None,
                 access_group_out=None, dhcp_relay_servers=None, arp_routing=None, icmp_redirects=None,
                 unicast_rpf_mode=None, ntp=None, varp_ips=None, load_interval=None, mpls_ip=None):
        self.number = number
        self.name = name
        self.access_groups = {IN: access_group_in, OUT: access_group_out}
        self.vrf_forwarding = vrf_forwarding
        self.ips = ips or []
        self.vrrp_groups = vrrp_groups or []
        self.dhcp_relay_servers = dhcp_relay_servers or []
        self.arp_routing = arp_routing
        self.icmp_redirects = icmp_redirects
        self.unicast_rpf_mode = unicast_rpf_mode
        self.ntp = ntp
        self.varp_ips = varp_ips or []
        self.load_interval = load_interval
        self.mpls_ip = mpls_ip



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from flask import request

from netman.api.api_utils import BadRequest, to_response
from netman.api.switch_api_base import SwitchApiBase
from netman.api.validators import resource, content, Session, \
    Resource, is_session


class SwitchSessionApi(SwitchApiBase):
    def __init__(self, switch_factory, sessions_manager):
        super(SwitchSessionApi, self).__init__(switch_factory, sessions_manager)
        self.server = None

    def hook_to(self, server):
        server.add_url_rule('/switches-sessions/<session_id>', view_func=self.open_session, methods=['POST'])
        server.add_url_rule('/switches-sessions/<session_id>', view_func=self.close_session, methods=['DELETE'])
        server.add_url_rule('/switches-sessions/<session_id>/actions', view_func=self.act_on_session, methods=['POST'])
        server.add_url_rule('/switches-sessions/<session_id>/<path:resource>', view_func=self.on_session, methods=['GET', 'PUT', 'POST', 'DELETE'])

        self.server = server
        return self

    @to_response
    @content(is_session)
    def open_session(self, session_id, hostname):
        """
        Open a locked session on a switch

        :arg str hostname: Hostname or IP of the switch
        :body:

        .. literalinclude:: ../doc_config/api_samples/post_switch_session.json
            :language: json

        :code 201 CREATED:

        Example output:

        .. literalinclude:: ../doc_config/api_samples/post_switch_session_result.json
            :language: json

        """

        switch = self.resolve_switch(hostname)
        session_id = self.sessions_manager.open_session(switch, session_id)

        return 201, {'session_id': session_id}

    @to_response
    @resource(Session)
    def close_session(self, session_id):
        """
        Close a session on a switch

        :arg str session: ID of the session

        """

        self.sessions_manager.close_session(session_id)

        return 204, None

    @to_response
    @resource(Session, Resource)
    def on_session(self, session_id, resource_name):
        self.sessions_manager.keep_alive(session_id)
        with self.server.test_client() as http_client:
            response = http_client.open(
                '/switches/{}/{}'.format(session_id, resource_name),
                method=request.method,
                headers={k: v for k, v in request.headers.items()},
                data=request.data)
            return response

    @to_response
    @resource(Session)
    def act_on_session(self, session_id):
        """
        Commit or rollback a session on a switch

        :arg str session: ID of the session
        :body:
            ``commit`` or ``rollback``
        """

        action = request.data.lower()
        if action == 'start_transaction':
            self.sessions_manager.start_transaction(session_id)
        elif action == 'end_transaction':
            self.sessions_manager.end_transaction(session_id)
        elif action == 'commit':
            self.sessions_manager.commit_session(session_id)
        elif action == 'rollback':
            self.sessions_manager.rollback_session(session_id)
        else:
            raise BadRequest('Unknown action {}'.format(action))

        return 204, None



import re
import warnings

import pyeapi
from netaddr import IPNetwork, IPAddress, AddrFormatError
from pyeapi.api.vlans import isvlan
from pyeapi.eapilib import CommandError

from netman import regex
from netman.adapters.shell import default_command_timeout
from netman.adapters.switches.util import split_on_dedent
from netman.api.validators import is_valid_mpls_state
from netman.core.objects.exceptions import VlanAlreadyExist, UnknownVlan, BadVlanNumber, BadVlanName, \
    IPAlreadySet, IPNotAvailable, UnknownIP, DhcpRelayServerAlreadyExists, UnknownDhcpRelayServer, UnknownInterface, \
    UnknownBond, VarpAlreadyExistsForVlan, VarpDoesNotExistForVlan, BadLoadIntervalNumber
from netman.core.objects.interface import Interface
from netman.core.objects.interface_states import OFF, ON
from netman.core.objects.port_modes import ACCESS, TRUNK
from netman.core.objects.switch_base import SwitchBase
from netman.core.objects.vlan import Vlan


def eapi_http(switch_descriptor):
    return Arista(switch_descriptor, transport="http")


def eapi_https(switch_descriptor):
    return Arista(switch_descriptor, transport="https")


def eapi(switch_descriptor):
    warnings.warn("Use either the _http or _https driver", DeprecationWarning)

    m = re.match('^(https?):\/\/(.*)$', switch_descriptor.hostname.lower())
    transport, hostname = (m.group(1), m.group(2)) if m else ('https', switch_descriptor.hostname)

    switch_descriptor.hostname = hostname
    return Arista(switch_descriptor, transport=transport)


class Arista(SwitchBase):
    def __init__(self, switch_descriptor, transport):
        super(Arista, self).__init__(switch_descriptor)
        self.switch_descriptor = switch_descriptor
        self.transport = transport

    def _connect(self):
        self.node = pyeapi.connect(host=self.switch_descriptor.hostname,
                                   username=self.switch_descriptor.username,
                                   password=self.switch_descriptor.password,
                                   port=self.switch_descriptor.port,
                                   transport=self.transport,
                                   return_node=True,
                                   timeout=default_command_timeout)

    def _disconnect(self):
        self.node = None

    def _end_transaction(self):
        pass

    def _start_transaction(self):
        pass

    def commit_transaction(self):
        self.node.enable('write memory')

    def rollback_transaction(self):
        pass

    def get_vlan(self, number):
        try:
            vlans_result, interfaces_result = self.node.enable(
                ["show vlan {}".format(number), "show interfaces Vlan{}".format(number)], strict=True)
            vlans_info = vlans_result['result']
            interfaces_info = interfaces_result['result']
        except CommandError as e:
            if "not found in current VLAN database" in e.command_error:
                raise UnknownVlan(number)
            elif "Invalid input" in e.command_error:
                raise BadVlanNumber()
            elif "Interface does not exist" in e.command_error:
                vlans_info = e.output[1]
                interfaces_info = {"interfaces": {}}
            else:
                raise

        vlans = _extract_vlans(vlans_info)
        _apply_interface_data(interfaces_info, vlans)
        self._apply_interface_vlan_data(vlans)

        return vlans[0]

    def get_vlans(self):
        vlans_result, interfaces_result = self.node.enable(["show vlan", "show interfaces"], strict=True)

        vlans = _extract_vlans(vlans_result['result'])
        _apply_interface_data(interfaces_result['result'], vlans)
        self._apply_interface_vlan_data(vlans)

        return sorted(vlans, key=lambda v: v.number)

    def add_vlan(self, number, name=None):
        if not isvlan(number):
            raise BadVlanNumber()
        try:
            self.node.enable(["show vlan {}".format(number)], strict=True)
            raise VlanAlreadyExist(number)
        except CommandError:
            pass

        commands = ["vlan {}".format(number)]
        if name is not None:
            commands.append("name {}".format(name))

        try:
            self.node.config(commands)
        except CommandError:
            raise BadVlanName()

    def remove_vlan(self, number):
        try:
            self.node.enable(["show vlan {}".format(number)], strict=True)
        except CommandError:
            raise UnknownVlan(number)

        self.node.config(["no interface Vlan{}".format(number), "no vlan {}".format(number)])

    def add_ip_to_vlan(self, vlan_number, ip_network):
        vlan = self.get_vlan(vlan_number)

        ip_found = next((ip for ip in vlan.ips if ip.ip == ip_network.ip), False)
        if ip_found:
            raise IPAlreadySet(ip_network, ip_found)

        has_ips = len(vlan.ips) > 0
        add_ip_command = "ip address {}{}".format(str(ip_network), " secondary" if has_ips else "")

        commands = [
            "interface vlan {}".format(vlan_number),
            add_ip_command
        ]
        try:
            self.node.config(commands)
        except CommandError as e:
            raise IPNotAvailable(ip_network, reason=str(e))

    def remove_ip_from_vlan(self, vlan_number, ip_network):
        vlan = self.get_vlan(vlan_number)
        existing_ip = next((ip for ip in vlan.ips
                            if ip.ip == ip_network.ip and ip.netmask == ip_network.netmask),
                           False)

        if existing_ip:
            ip_index = vlan.ips.index(existing_ip)
            if ip_index == 0:
                if len(vlan.ips) == 1:
                    remove_ip_command = "no ip address {}".format(str(ip_network))
                else:
                    remove_ip_command = "ip address {}".format(str(vlan.ips[1]))
            else:
                remove_ip_command = "no ip address {} secondary".format(str(ip_network))

            commands = [
                "interface Vlan{}".format(vlan_number),
                remove_ip_command
            ]
            self.node.config(commands)
        else:
            raise UnknownIP(ip_network)

    def get_interface(self, interface_id):
        commands = [
            "show interfaces {}".format(interface_id),
            "show interfaces {} switchport".format(interface_id)
        ]
        try:
            result = self.node.enable(commands, strict=True)
        except CommandError:
            raise UnknownInterface(interface_id)

        interfaces = parse_interfaces(result[0]['result']['interfaces'], result[1]['result']['switchports'])
        if len(interfaces) > 0:
            return interfaces[0]
        raise UnknownInterface(interface_id)

    def get_interfaces(self):
        commands = [
            "show interfaces",
            "show interfaces switchport"
        ]
        result = self.node.enable(commands, strict=True)

        interfaces = parse_interfaces(result[0]['result']['interfaces'], result[1]['result']['switchports'])
        return interfaces

    def set_trunk_mode(self, interface_id):
        commands = [
            "interface {}".format(interface_id),
            "switchport mode trunk",
            "switchport trunk allowed vlan none"
        ]
        try:
            self.node.config(commands)
        except CommandError:
            raise UnknownInterface(interface_id)

    def add_trunk_vlan(self, interface_id, vlan):
        self.get_vlan(vlan)
        commands = [
            "interface {}".format(interface_id),
            "switchport trunk allowed vlan add {}".format(vlan)
        ]
        try:
            self.node.config(commands)
        except CommandError:
            raise UnknownInterface(interface_id)

    def remove_trunk_vlan(self, interface_id, vlan):
        interface = self.get_interface(interface_id)
        if vlan not in interface.trunk_vlans:
            raise UnknownVlan(vlan)

        commands = [
            "interface {}".format(interface_id),
            "switchport trunk allowed vlan remove {}".format(vlan)
        ]
        self.node.config(commands)

    def set_bond_trunk_mode(self, number):
        with NamedBond(number) as bond:
            return self.set_trunk_mode(bond.name)

    def add_bond_trunk_vlan(self, number, vlan):
        with NamedBond(number) as bond:
            return self.add_trunk_vlan(bond.name, vlan)

    def remove_bond_trunk_vlan(self, number, vlan):
        with NamedBond(number) as bond:
            return self.remove_trunk_vlan(bond.name, vlan)

    def add_dhcp_relay_server(self, vlan_number, ip_address):
        vlan = self.get_vlan(vlan_number)

        if ip_address in vlan.dhcp_relay_servers:
            raise DhcpRelayServerAlreadyExists(vlan_number=vlan_number, ip_address=ip_address)

        self.node.config(['interface Vlan{}'.format(vlan_number),
                          'ip helper-address {}'.format(ip_address)])

    def remove_dhcp_relay_server(self, vlan_number, ip_address):
        vlan = self.get_vlan(vlan_number)

        if ip_address not in vlan.dhcp_relay_servers:
            raise UnknownDhcpRelayServer(vlan_number=vlan_number, ip_address=ip_address)

        self.node.config(['interface Vlan{}'.format(vlan_number),
                          'no ip helper-address {}'.format(ip_address)])

    def set_vlan_load_interval(self, vlan_number, time_interval):
        self.get_vlan(vlan_number)

        try:
            self.node.config(['interface Vlan{}'.format(vlan_number),
                              'load-interval {}'.format(time_interval)])
        except CommandError:
            raise BadLoadIntervalNumber()

    def unset_vlan_load_interval(self, vlan_number):
        self.get_vlan(vlan_number)

        self.node.config(['interface Vlan{}'.format(vlan_number),
                          'no load-interval'])

    def set_vlan_mpls_ip_state(self, vlan_number, state):
        is_valid_mpls_state(state)
        self.get_vlan(vlan_number)

        self.node.config(['interface Vlan{}'.format(vlan_number),
                          'mpls ip' if state else 'no mpls ip'])

    def add_vlan_varp_ip(self, vlan_number, ip_network):
        vlan = self.get_vlan(vlan_number)

        if ip_network in vlan.varp_ips:
            raise VarpAlreadyExistsForVlan(vlan=vlan_number, ip_network=ip_network)

        try:
            self.node.config(['interface Vlan{}'.format(vlan_number),
                              'ip virtual-router address {}'.format(ip_network)])
        except CommandError as e:
            if regex.match("^.*is already assigned to interface Vlan(\d+)]", e.message):
                raise IPNotAvailable(ip_network=ip_network, reason=str(e))
            raise

    def remove_vlan_varp_ip(self, vlan_number, ip_network):
        vlan = self.get_vlan(vlan_number)

        if ip_network not in vlan.varp_ips:
            raise VarpDoesNotExistForVlan(vlan=vlan_number, ip_network=ip_network)

        self.node.config(['interface Vlan{}'.format(vlan_number),
                          'no ip virtual-router address {}'.format(ip_network)])

    def _apply_interface_vlan_data(self, vlans):
        config = self._fetch_interface_vlans_config(vlans)

        for interface in split_on_dedent(config):
            if regex.match("^.*Vlan(\d+)$", interface[0]):
                vlan = _find_vlan_by_number(vlans, regex[0])
                for line in interface[1:]:
                    if regex.match(" *ip helper-address (.*)", line):
                        try:
                            vlan.dhcp_relay_servers.append(IPAddress(regex[0]))
                        except AddrFormatError:
                            self.logger.warning(
                                'Unsupported IP Helper address found in Vlan {} : {}'.format(vlan.number, regex[0]))
                    if regex.match(" *ip virtual-router address (.*)", line):
                        vlan.varp_ips.append(IPNetwork(regex[0]))
                    if regex.match(" *load-interval (.*)", line):
                        vlan.load_interval = int(regex[0])
                    if regex.match(" *no mpls ip", line):
                        vlan.mpls_ip = False

    def _fetch_interface_vlans_config(self, vlans):
        all_interface_vlans = sorted('Vlan{}'.format(vlan.number) for vlan in vlans)
        return self.node.get_config(params='interfaces {}'.format(' '.join(all_interface_vlans)))


def _find_vlan_by_number(vlans, number):
    return next((vlan for vlan in vlans if vlan.number == int(number)))


def parse_interfaces(interfaces_data, switchports_data):
    interfaces = []
    for interface_data in interfaces_data.values():
        if regex.match("(\w*Ethernet[^\s]*)", interface_data["name"]) or \
                regex.match("(Port-channel[^\s]*)", interface_data["name"]):

            interface = Interface(name=interface_data["name"], shutdown=False)

            if interface_data["lineProtocolStatus"] == "down":
                interface.shutdown = True

            interface.mtu = int(interface_data["mtu"])
            interface.auto_negotiation = ON if interface_data["autoNegotiate"] == "on" else OFF

            if interface.name in switchports_data:
                patch_switchport(interface, switchports_data[interface.name]["switchportInfo"])

            interfaces.append(interface)

    return interfaces


def patch_switchport(interface, data):
    if data["mode"] == "access":
        interface.port_mode = ACCESS
    elif data["mode"] == "trunk":
        interface.port_mode = TRUNK

    interface.trunk_native_vlan = data["trunkingNativeVlanId"]
    interface.trunk_vlans = parse_vlan_ranges(data["trunkAllowedVlans"]) if data["trunkAllowedVlans"] else []


def parse_vlan_ranges(all_ranges):
    if all_ranges is None or all_ranges == "ALL":
        return range(1, 4094)
    elif all_ranges == "NONE":
        return []
    else:
        full_list = []
        for vlan_list in [parse_range(r) for r in all_ranges.split(",")]:
            full_list += vlan_list
        return full_list


def parse_range(single_range):
    if regex.match("(\d+)-(\d+)", single_range):
        return range(int(regex[0]), int(regex[1]) + 1)
    else:
        return [int(single_range)]


def bond_name(number):
    return "Port-Channel{}".format(number)


class NamedBond(object):
    def __init__(self, number):
        self.number = number

    @property
    def name(self):
        return bond_name(self.number)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is UnknownInterface:
            raise UnknownBond(self.number), None, exc_tb


def _extract_vlans(vlans_info):
    vlan_list = []
    for id, vlan in vlans_info['vlans'].items():
        if vlan['name'] == "VLAN{:04d}".format(int(id)):
            vlan['name'] = None

        vlan_list.append(Vlan(number=int(id), name=vlan['name'], icmp_redirects=True,
                              arp_routing=True, ntp=True, mpls_ip=True))
    return vlan_list


def _apply_interface_data(interfaces_result, vlans):
    interfaces = interfaces_result["interfaces"]
    for vlan in vlans:
        interface_name = "Vlan{}".format(vlan.number)

        interface_data = interfaces.get(interface_name)
        if interface_data is not None and len(interface_data["interfaceAddress"]) > 0:
            interface_address = interface_data["interfaceAddress"][0]

            primary_ip_data = interface_address["primaryIp"]
            secondary_ips_data = interface_address["secondaryIpsOrderedList"]

            primary_ip = _to_ip(primary_ip_data)

            if primary_ip.ip != IPNetwork("0.0.0.0/0").ip:
                vlan.ips.append(primary_ip)

                for addr in secondary_ips_data:
                    vlan.ips.append(_to_ip(addr))


def _to_ip(ip_data):
    return IPNetwork("{}/{}".format(ip_data["address"], ip_data["maskLen"]))



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from hamcrest import assert_that, is_

from netman.core.objects.exceptions import UnknownInterface
from netman.core.objects.interface_states import ON, OFF
from tests import has_message
from tests.adapters.compliance_test_case import ComplianceTestCase


class SetInterfaceStateTest(ComplianceTestCase):
    _dev_sample = "juniper"

    def test_can_open_an_interface(self):
        self.client.set_interface_state(self.test_port, state=ON)
        interface = self.client.get_interface(self.test_port)

        assert_that(interface.shutdown, is_(False))

    def test_can_shutdown_an_interface(self):
        self.client.set_interface_state(self.test_port, state=OFF)
        interface = self.client.get_interface(self.test_port)

        assert_that(interface.shutdown, is_(True))

    def test_fails_if_the_interface_does_not_exist(self):
        with self.assertRaises(UnknownInterface) as expect:
            self.client.set_interface_state('ge-0/0/128', state=ON)

        assert_that(expect.exception, has_message("Unknown interface ge-0/0/128"))

    def test_set_an_interface_twice_works(self):
        self.try_to.set_interface_state(self.test_port, ON)
        self.try_to.set_interface_state(self.test_port, ON)

        self.try_to.set_interface_state(self.test_port, OFF)
        self.try_to.set_interface_state(self.test_port, OFF)

    def tearDown(self):
        self.janitor.unset_interface_state(self.test_port)
        super(SetInterfaceStateTest, self).tearDown()



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json


class FlaskResponse(object):
    def __init__(self, response):
        self.response = response
        self.status_code = response.status_code
        self.data = response.data
        self.headers = response.headers

    def json(self):
        return json.loads(self.response.data)


class FlaskRequest(object):
    def __init__(self, flask_client):
        self.flask_client = flask_client

    def get(self, url, **kwargs):
        return FlaskResponse(self.flask_client.get(path=url, **kwargs))

    def post(self, url, **kwargs):
        return FlaskResponse(self.flask_client.post(path=url, **kwargs))

    def put(self, url, **kwargs):
        return FlaskResponse(self.flask_client.put(path=url, **kwargs))

    def delete(self, url, **kwargs):
        return FlaskResponse(self.flask_client.delete(path=url, **kwargs))



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from _socket import timeout, gaierror
import logging
import time

import paramiko
from netman.adapters import shell

from netman.adapters.shell.base import TerminalClient
from netman.core.objects.exceptions import CouldNotConnect, ConnectTimeout, CommandTimeout


class SshClient(TerminalClient):

    def __init__(self, host, username, password, port=22, prompt=('>', '#'), connect_timeout=None, command_timeout=None,
                 reading_interval=0.01, reading_chunk_size=9999):
        self.logger = logging.getLogger(__name__)

        self.host = host
        self.port = port
        self.username = username
        self.prompt = prompt
        self.command_timeout = command_timeout or shell.default_command_timeout
        connect_timeout = connect_timeout or shell.default_connect_timeout
        self.reading_interval = reading_interval
        self.reading_chunk_size = reading_chunk_size

        self.current_buffer = ''
        self.client = None
        self.channel = None
        self.full_log = ""

        self._open_channel(host, port, username, password, connect_timeout)

    def do(self, command, wait_for=None, include_last_line=False):
        self.logger.debug("[SSH][{}@{}:{}] Send >> {}".format(self.username, self.host, self.port, command))

        self.channel.send(command + '\n')
        return self._read_until(wait_for, include_last_line)

    def send_key(self, key, wait_for=None, include_last_line=False):
        self.logger.debug("[SSH][{}@{}:{}] Send KEY >> {}".format(self.username, self.host, self.port, key))

        self.channel.send(key)
        return self._read_until(wait_for, include_last_line)

    def quit(self, command):
        self.logger.debug("[SSH][{}@{}:{}] Quit >> {}".format(self.username, self.host, self.port, command))

        self.channel.send(command + '\n')

    def get_current_prompt(self):
        return self.current_buffer.splitlines()[-1]

    def _open_channel(self, host, port, username, password, connect_timeout):
        self.client = paramiko.SSHClient()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        try:
            self.client.connect(host,
                                port=port,
                                username=username,
                                password=password,
                                timeout=connect_timeout,
                                allow_agent=False,
                                look_for_keys=False)
        except timeout:
            raise ConnectTimeout(host, port)
        except gaierror:
            raise CouldNotConnect(host, port)

        self.channel = self.client.invoke_shell()

        self._wait_for(self.prompt)

    def _read_until(self, wait_for, include_last_line):
        self._wait_for(wait_for or self.prompt)

        lines = self.current_buffer.splitlines()[1:]
        if not include_last_line:
            lines = lines[:-1]

        return filter(None, lines)

    def _wait_for(self, wait_for):
        self.current_buffer = ''

        started_at = time.time()
        while not self.current_buffer.endswith(wait_for):
            while not self.channel.recv_ready():
                if time.time() - started_at > self.command_timeout:
                    raise CommandTimeout(wait_for, self.current_buffer)
                time.sleep(self.reading_interval)

            read = self.channel.recv(self.reading_chunk_size)
            self.logger.debug("[SSH][{}@{}:{}] Recv << {}".format(self.username, self.host, self.port, repr(read)))
            self.full_log += read
            self.current_buffer += read



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest

from hamcrest import assert_that, equal_to, is_, instance_of
import mock
from ncclient.operations import RPCError
from netaddr import IPAddress, IPNetwork
from flexmock import flexmock, flexmock_teardown

from netman.core.objects.interface_states import OFF, ON
from tests import ExactIpNetwork, ignore_deprecation_warnings
from tests.api import open_fixture
from netman.adapters.switches.remote import RemoteSwitch, factory
from netman.core.objects.access_groups import IN, OUT
from netman.core.objects.exceptions import UnknownBond, VlanAlreadyExist, BadBondLinkSpeed, LockedSwitch, \
    NetmanException, UnknownInterface, UnknownSession, UnknownVlan, BadMplsIpState
from netman.core.objects.port_modes import ACCESS, TRUNK, DYNAMIC
from netman.core.objects.switch_descriptor import SwitchDescriptor


class AnException(Exception):
    pass


@ignore_deprecation_warnings
def test_factory():
    switch = factory(SwitchDescriptor(hostname='hostname', model='juniper', username='username', password='password', port=22))

    assert_that(switch, instance_of(RemoteSwitch))
    assert_that(switch.switch_descriptor.hostname, equal_to("hostname"))
    assert_that(switch.switch_descriptor.model, equal_to("juniper"))
    assert_that(switch.switch_descriptor.username, equal_to("username"))
    assert_that(switch.switch_descriptor.password, equal_to("password"))
    assert_that(switch.switch_descriptor.port, equal_to(22))


class RemoteSwitchTest(unittest.TestCase):
    netman_url = 'http://netman.example.org:1234'

    def setUp(self):
        self.switch = RemoteSwitch(SwitchDescriptor(
            model="juniper", hostname="toto", username="tutu",
            password="titi", port=1234, netman_server=self.netman_url))

        self.requests_mock = flexmock()
        self.switch.requests = self.requests_mock
        self.headers = {
            'Netman-Port': "1234",
            'Netman-Model': 'juniper',
            'Netman-Password': 'titi',
            'Netman-Username': 'tutu',
            'Netman-Max-Version': "2",
            'Netman-Verbose-Errors': 'yes',
        }

    def tearDown(self):
        flexmock_teardown()

    def test_switch_has_a_logger_configured_with_the_switch_name(self):
        assert_that(self.switch.logger.name, is_(RemoteSwitch.__module__ + ".toto"))

    @mock.patch('uuid.uuid4')
    def test_start_then_commit_returns_to_normal_behavior(self, m_uuid):
        m_uuid.return_value = '0123456789'
        self.headers['Netman-Session-Id'] = '0123456789'
        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches-sessions/0123456789',
            headers=self.headers,
            data=JsonData(hostname="toto")
        ).and_return(
            Reply(
                content=json.dumps({'session_id': '0123456789'}),
                status_code=201))
        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches-sessions/0123456789/actions',
            data='start_transaction',
            headers={'Netman-Verbose-Errors': "yes",
                     'Netman-Max-Version': "2",
                     'Netman-Session-Id': '0123456789'}
        ).and_return(
            Reply(
                content="",
                status_code=204))
        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches-sessions/0123456789/actions',
            data='commit',
            headers={'Netman-Verbose-Errors': "yes",
                     'Netman-Max-Version': "2",
                     'Netman-Session-Id': '0123456789'}
        ).and_return(
            Reply(
                content="",
                status_code=204))
        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches-sessions/0123456789/actions',
            data='end_transaction',
            headers={'Netman-Verbose-Errors': "yes",
                     'Netman-Max-Version': "2",
                     'Netman-Session-Id': '0123456789'}
        ).and_return(
            Reply(
                content="",
                status_code=204))
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches-sessions/0123456789',
            headers={'Netman-Verbose-Errors': "yes",
                     'Netman-Max-Version': "2",
                     'Netman-Session-Id': '0123456789'}
        ).and_return(
            Reply(
                content="",
                status_code=204))

        self.switch.connect()
        self.switch.start_transaction()
        self.switch.commit_transaction()
        self.switch.end_transaction()
        self.switch.disconnect()
        self.setUp()
        self.test_add_bond()

    @mock.patch('uuid.uuid4')
    def test_connect_fails_to_obtain_a_session(self, m_uuid):
        m_uuid.return_value = '0123456789'
        self.headers['Netman-Session-Id'] = '0123456789'
        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches-sessions/0123456789',
            headers=self.headers,
            data=JsonData(hostname="toto")
        ).and_return(
            Reply(
                content=json.dumps({
                    "error": "",
                    "error-module": AnException.__module__,
                    "error-class": AnException.__name__
                }),
                status_code=500))

        with self.assertRaises(AnException):
            self.switch.connect()

    @mock.patch('uuid.uuid4')
    def test_disconnect_fails_and_return_to_normal_behavior(self, m_uuid):
        m_uuid.return_value = '0123456789'
        self.headers['Netman-Session-Id'] = '0123456789'
        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches-sessions/0123456789',
            headers=self.headers,
            data=JsonData(hostname="toto")
        ).and_return(
            Reply(
                content=json.dumps({'session_id': '0123456789'}),
                status_code=201))

        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches-sessions/0123456789',
            headers={'Netman-Verbose-Errors': "yes",
                     'Netman-Max-Version': "2",
                     'Netman-Session-Id': '0123456789'}
        ).and_return(
            Reply(
                content=json.dumps({
                    "error": "",
                    "error-module": AnException.__module__,
                    "error-class": AnException.__name__
                }),
                status_code=500))

        self.switch.connect()
        with self.assertRaises(AnException):
            self.switch.disconnect()
        self.setUp()
        self.test_add_bond()

    @mock.patch('uuid.uuid4')
    def test_session_is_used_when_we_are_in_a_session(self, m_uuid):
        m_uuid.return_value = '0123456789'
        self.headers['Netman-Session-Id'] = '0123456789'
        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches-sessions/0123456789',
            headers=self.headers,
            data=JsonData(hostname="toto")
        ).and_return(
            Reply(
                content=json.dumps({'session_id': '0123456789'}),
                status_code=201))

        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches-sessions/0123456789/bonds',
            headers=self.headers,
            data=JsonData(number=6)
        ).and_return(
            Reply(
                content='',
                status_code=201))

        self.switch.connect()
        self.switch.add_bond(6)

    @mock.patch('uuid.uuid4')
    def test_commit_transaction_fails_to_commit(self, m_uuid):
        m_uuid.return_value = '0123456789'

        self.headers['Netman-Session-Id'] = '0123456789'
        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches-sessions/0123456789',
            headers=self.headers,
            data=JsonData(hostname="toto")
        ).and_return(
            Reply(
                content=json.dumps({'session_id': '0123456789'}),
                status_code=201))

        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches-sessions/0123456789/actions',
            data='commit',
            headers={'Netman-Verbose-Errors': "yes",
                     'Netman-Max-Version': "2",
                     'Netman-Session-Id': '0123456789'}
        ).and_return(
            Reply(
                content=json.dumps({
                    "error": "",
                    "error-module": AnException.__module__,
                    "error-class": AnException.__name__
                }),
                status_code=500))

        self.switch.connect()
        with self.assertRaises(AnException):
            self.switch.commit_transaction()

    @mock.patch('uuid.uuid4')
    def test_rollback_transaction_fails_to_rollback(self, m_uuid):
        m_uuid.return_value = '0123456789'
        self.headers['Netman-Session-Id'] = '0123456789'
        self.requests_mock.should_receive("post").once().ordered().with_args(
            url=self.netman_url+'/switches-sessions/0123456789',
            headers=self.headers,
            data=JsonData(hostname="toto")
        ).and_return(
            Reply(
                content=json.dumps({'session_id': '0123456789'}),
                status_code=201))

        self.requests_mock.should_receive("post").once().ordered().with_args(
            url=self.netman_url+'/switches-sessions/0123456789/actions',
            data='rollback',
            headers={'Netman-Verbose-Errors': "yes",
                     'Netman-Max-Version': "2",
                     'Netman-Session-Id': '0123456789'}
        ).and_return(
            Reply(
                content=json.dumps({
                    "error": "",
                    "error-module": AnException.__module__,
                    "error-class": AnException.__name__
                }),
                status_code=500))

        self.switch.connect()
        with self.assertRaises(AnException):
            self.switch.rollback_transaction()

    @mock.patch('uuid.uuid4')
    def test_receiving_unknown_session_during_transaction_will_connect_again(self, m_uuid):
        m_uuid.return_value = '0123456789'
        first_connect_headers = self.headers.copy()
        first_connect_headers['Netman-Session-Id'] = '0123456789'

        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches-sessions/0123456789',
            headers=first_connect_headers,
            data=JsonData(hostname="toto")
        ).and_return(
            Reply(
                content=json.dumps({'session_id': '0123456789'}),
                status_code=201))

        self.switch.connect()

        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches-sessions/0123456789/bonds',
            headers=first_connect_headers,
            data=JsonData(number=6)
        ).and_return(
            Reply(
                content=json.dumps({
                    "error": "",
                    "error-module": UnknownSession.__module__,
                    "error-class": UnknownSession.__name__
                }),
                status_code=500))

        m_uuid.return_value = 'new-session-id'
        second_connect_headers = self.headers.copy()
        second_connect_headers['Netman-Session-Id'] = 'new-session-id'

        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches-sessions/new-session-id',
            headers=second_connect_headers,
            data=JsonData(hostname="toto")
        ).and_return(
            Reply(
                content=json.dumps({'session_id': 'new-session-id'}),
                status_code=201))

        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches-sessions/new-session-id/bonds',
            headers=second_connect_headers,
            data=JsonData(number=6)
        ).and_return(
            Reply(
                content='',
                status_code=201))

        self.switch.add_bond(6)

    @mock.patch('uuid.uuid4')
    def test_receiving_unknown_session_twice_during_transaction_will_raise_an_exception(self, m_uuid):
        m_uuid.return_value = '0123456789'
        first_connect_headers = self.headers.copy()
        first_connect_headers['Netman-Session-Id'] = '0123456789'

        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches-sessions/0123456789',
            headers=first_connect_headers,
            data=JsonData(hostname="toto")
        ).and_return(
            Reply(
                content=json.dumps({'session_id': '0123456789'}),
                status_code=201))

        self.switch.connect()

        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches-sessions/0123456789/bonds',
            headers=first_connect_headers,
            data=JsonData(number=6)
        ).and_return(
            Reply(
                content=json.dumps({
                    "error": "",
                    "error-module": UnknownSession.__module__,
                    "error-class": UnknownSession.__name__
                }),
                status_code=500))

        m_uuid.return_value = 'new-session-id'
        second_connect_headers = self.headers.copy()
        second_connect_headers['Netman-Session-Id'] = 'new-session-id'

        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches-sessions/new-session-id',
            headers=second_connect_headers,
            data=JsonData(hostname="toto")
        ).and_return(
            Reply(
                content=json.dumps({'session_id': 'new-session-id'}),
                status_code=201))

        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches-sessions/new-session-id/bonds',
            headers=second_connect_headers,
            data=JsonData(number=6)
        ).and_return(
            Reply(
                content=json.dumps({
                    "error": "",
                    "error-module": UnknownSession.__module__,
                    "error-class": UnknownSession.__name__
                }),
                status_code=500))

        with self.assertRaises(UnknownSession):
            self.switch.add_bond(6)

    @mock.patch('uuid.uuid4')
    def test_multi_proxy_1(self, m_uuid):
        self.switch = RemoteSwitch(SwitchDescriptor(
            model="juniper", hostname="toto", username="tutu",
            password="titi", port=1234, netman_server=[self.netman_url, "1.2.3.4"]))

        self.requests_mock = flexmock()
        self.switch.requests = self.requests_mock

        m_uuid.return_value = '0123456789'
        self.requests_mock.should_receive("post").once().ordered().with_args(
            url=self.netman_url+'/switches-sessions/0123456789',
            headers={
                'Netman-Port': "1234",
                'Netman-Model': 'juniper',
                'Netman-Password': 'titi',
                'Netman-Username': 'tutu',
                'Netman-Verbose-Errors': 'yes',
                'Netman-Proxy-Server': '1.2.3.4',
                'Netman-Max-Version': "2",
                'Netman-Session-Id': '0123456789'
            },
            data=JsonData(hostname="toto")
        ).and_return(
            Reply(
                content=json.dumps({'session_id': '0123456789'}),
                status_code=201))

        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches-sessions/0123456789',
            headers={'Netman-Verbose-Errors': "yes",
                     'Netman-Max-Version': "2",
                     'Netman-Session-Id': '0123456789'}
        ).and_return(
            Reply(
                content="",
                status_code=204))

        self.switch.connect()
        self.switch.disconnect()
        self.setUp()
        self.test_add_bond()

    @mock.patch('uuid.uuid4')
    def test_multi_proxy_few(self, m_uuid):
        self.switch = RemoteSwitch(SwitchDescriptor(
            model="juniper", hostname="toto", username="tutu",
            password="titi", port=1234, netman_server=[self.netman_url, "1.2.3.4", "5.6.7.8"]))

        self.requests_mock = flexmock()
        self.switch.requests = self.requests_mock

        m_uuid.return_value = '0123456789'
        self.requests_mock.should_receive("post").once().ordered().with_args(
            url=self.netman_url+'/switches-sessions/0123456789',
            headers={
                'Netman-Port': "1234",
                'Netman-Model': 'juniper',
                'Netman-Password': 'titi',
                'Netman-Username': 'tutu',
                'Netman-Verbose-Errors': 'yes',
                'Netman-Proxy-Server': '1.2.3.4,5.6.7.8',
                'Netman-Max-Version': "2",
                'Netman-Session-Id': '0123456789'
            },
            data=JsonData(hostname="toto")
        ).and_return(
            Reply(
                content=json.dumps({'session_id': '0123456789'}),
                status_code=201))

        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches-sessions/0123456789',
            headers={'Netman-Verbose-Errors': "yes",
                     'Netman-Max-Version': "2",
                     'Netman-Session-Id': '0123456789'}
        ).and_return(
            Reply(
                content="",
                status_code=204))

        self.switch.connect()
        self.switch.disconnect()
        self.setUp()
        self.test_add_bond()

    def test_get_vlan(self):
        self.requests_mock.should_receive("get").once().with_args(
            url=self.netman_url+'/switches/toto/vlans/1',
            headers=self.headers
        ).and_return(
            Reply(
                content=open_fixture('get_switch_hostname_vlans_vlan.json').read(),
                status_code=200))

        vlan1 = self.switch.get_vlan(1)

        assert_that(vlan1.number, is_(1))
        assert_that(vlan1.name, is_('One'))
        assert_that(vlan1.ips, is_([ExactIpNetwork('1.1.1.1', 24)]))
        assert_that(vlan1.vrf_forwarding, is_("MY_VRF"))
        assert_that(vlan1.access_groups[IN], is_("Blah_blah"))
        assert_that(vlan1.access_groups[OUT], is_(None))
        assert_that(vlan1.dhcp_relay_servers, is_([]))
        vrrp_group = vlan1.vrrp_groups[0]
        assert_that(vrrp_group.id, is_(1))
        assert_that(vrrp_group.ips, is_([IPAddress("1.1.1.2")]))
        assert_that(vrrp_group.priority, is_(90))
        assert_that(vrrp_group.hello_interval, is_(5))
        assert_that(vrrp_group.dead_interval, is_(15))
        assert_that(vrrp_group.track_id, is_("101"))
        assert_that(vrrp_group.track_decrement, is_(50))
        assert_that(vlan1.varp_ips, is_([]))

    def test_get_vlans(self):
        self.requests_mock.should_receive("get").once().with_args(
            url=self.netman_url+'/switches/toto/vlans',
            headers=self.headers
        ).and_return(
            Reply(
                content=open_fixture('get_switch_hostname_vlans.json').read(),
                status_code=200))

        vlan1, vlan2 = self.switch.get_vlans()

        assert_that(vlan1.number, is_(1))
        assert_that(vlan1.name, is_('One'))
        assert_that(vlan1.ips, is_([ExactIpNetwork('1.1.1.1', 24)]))
        assert_that(vlan1.vrf_forwarding, is_("MY_VRF"))
        assert_that(vlan1.access_groups[IN], is_("Blah_blah"))
        assert_that(vlan1.access_groups[OUT], is_(None))
        assert_that(vlan1.dhcp_relay_servers, is_([]))
        vrrp_group = vlan1.vrrp_groups[0]
        assert_that(vrrp_group.id, is_(1))
        assert_that(vrrp_group.ips, is_([IPAddress("1.1.1.2")]))
        assert_that(vrrp_group.priority, is_(90))
        assert_that(vrrp_group.hello_interval, is_(5))
        assert_that(vrrp_group.dead_interval, is_(15))
        assert_that(vrrp_group.track_id, is_("101"))
        assert_that(vrrp_group.track_decrement, is_(50))
        assert_that(vlan1.varp_ips, is_([IPNetwork('3.3.3.3/24')]))

        assert_that(vlan2.number, is_(2))
        assert_that(vlan2.name, is_(''))
        assert_that(vlan2.ips, is_([ExactIpNetwork('2.2.2.2', 24), ExactIpNetwork('3.3.3.3', 24)]))
        assert_that(vlan2.vrf_forwarding, is_(None))
        assert_that(vlan2.access_groups[IN], is_(None))
        assert_that(vlan2.access_groups[OUT], is_(None))
        assert_that(vlan2.dhcp_relay_servers, is_([IPAddress("10.10.10.1")]))
        vrrp_group1, vrrp_group2 = vlan2.vrrp_groups
        assert_that(vrrp_group1.id, is_(1))
        assert_that(vrrp_group1.ips, is_([IPAddress("2.2.2.2")]))
        assert_that(vrrp_group1.priority, is_(100))
        assert_that(vrrp_group1.hello_interval, is_(None))
        assert_that(vrrp_group1.dead_interval, is_(None))
        assert_that(vrrp_group1.track_id, is_(None))
        assert_that(vrrp_group1.track_decrement, is_(None))
        assert_that(vrrp_group2.id, is_(2))
        assert_that(vrrp_group2.ips, is_([IPAddress("3.3.3.1")]))
        assert_that(vrrp_group2.priority, is_(100))
        assert_that(vlan2.varp_ips, is_([IPNetwork('4.4.4.4/24'), IPNetwork('5.5.5.5/24')]))

    def test_get_vlan_interfaces(self):
        self.requests_mock.should_receive("get").once().with_args(
                url=self.netman_url+'/switches/toto/vlans/1/interfaces',
                headers=self.headers
        ).and_return(
            Reply(
                content=open_fixture('get_switch_hostname_vlans_vlan_interfaces.json').read(),
                status_code=200)
        )

        interfaces = self.switch.get_vlan_interfaces(1)

        assert_that(interfaces, is_(["ethernet 1/4", "FastEthernet0/3", "GigabitEthernet0/8"]))

    def test_get_vlan_interfaces_with_no_vlan_raises(self):
        self.requests_mock.should_receive("get").once().with_args(
                url=self.netman_url+'/switches/toto/vlans/4000/interfaces',
                headers=self.headers
        ).and_return(
                Reply(
                        content=json.dumps({
                            "error": "Vlan 4000 not found",
                            "error-module": UnknownVlan.__module__,
                            "error-class": UnknownVlan.__name__
                        }),
                        status_code=404))

        with self.assertRaises(UnknownVlan) as expect:
            self.switch.get_vlan_interfaces('4000')

        assert_that(str(expect.exception), equal_to("Vlan 4000 not found"))

    def test_get_interface(self):
        self.requests_mock.should_receive("get").once().with_args(
                url=self.netman_url+'/switches/toto/interfaces/ethernet 1/4',
                headers=self.headers
        ).and_return(
                Reply(
                        content=open_fixture('get_switch_hostname_interface.json').read(),
                        status_code=200))

        interface = self.switch.get_interface('ethernet 1/4')

        assert_that(interface.name, equal_to("ethernet 1/4"))
        assert_that(interface.shutdown, equal_to(False))
        assert_that(interface.port_mode, equal_to(TRUNK))
        assert_that(interface.access_vlan, equal_to(None))
        assert_that(interface.trunk_native_vlan, equal_to(2999))
        assert_that(interface.trunk_vlans, equal_to([3000, 3001, 3002]))

    def test_get_nonexistent_interface_raises(self):
        self.requests_mock.should_receive("get").once().with_args(
                url=self.netman_url+'/switches/toto/interfaces/ethernet 1/INEXISTENT',
                headers=self.headers
        ).and_return(
                Reply(
                        content=json.dumps({
                            "error": "Interface ethernet 1/INEXISTENT not found",
                            "error-module": UnknownInterface.__module__,
                            "error-class": UnknownInterface.__name__
                        }),
                        status_code=404))

        with self.assertRaises(UnknownInterface) as expect:
            self.switch.get_interface('ethernet 1/INEXISTENT')

        assert_that(str(expect.exception), equal_to("Interface ethernet 1/INEXISTENT not found"))

    def test_get_interfaces(self):
        self.requests_mock.should_receive("get").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces',
            headers=self.headers
        ).and_return(
            Reply(
                content=open_fixture('get_switch_hostname_interfaces.json').read(),
                status_code=200))

        if1, if2, if3, if4 = self.switch.get_interfaces()

        assert_that(if1.name, equal_to("ethernet 1/4"))
        assert_that(if1.shutdown, equal_to(False))
        assert_that(if1.port_mode, equal_to(TRUNK))
        assert_that(if1.access_vlan, equal_to(None))
        assert_that(if1.trunk_native_vlan, equal_to(2999))
        assert_that(if1.trunk_vlans, equal_to([3000, 3001, 3002]))

        assert_that(if2.name, equal_to("FastEthernet0/3"))
        assert_that(if2.shutdown, equal_to(True))
        assert_that(if2.port_mode, equal_to(ACCESS))
        assert_that(if2.access_vlan, equal_to(1999))
        assert_that(if2.trunk_native_vlan, equal_to(None))
        assert_that(if2.trunk_vlans, equal_to([]))

        assert_that(if3.name, equal_to("GigabitEthernet0/6"))
        assert_that(if3.port_mode, equal_to(DYNAMIC))
        assert_that(if3.access_vlan, equal_to(1999))
        assert_that(if3.trunk_native_vlan, equal_to(2999))
        assert_that(if3.trunk_vlans, equal_to([3000, 3001, 3002]))

        assert_that(if4.name, equal_to("GigabitEthernet0/8"))
        assert_that(if4.shutdown, equal_to(False))
        assert_that(if4.bond_master, equal_to(12))

    @ignore_deprecation_warnings
    def test_get_bond_v1(self):
        self.requests_mock.should_receive("get").once().with_args(
            url=self.netman_url+'/switches/toto/bonds/3',
            headers=self.headers
        ).and_return(
            Reply(
                content=open_fixture('get_switch_hostname_bond_v1.json').read(),
                status_code=200))

        if1 = self.switch.get_bond(3)

        assert_that(if1.number, equal_to(3))
        assert_that(if1.link_speed, equal_to('1g'))
        assert_that(if1.interface.name, equal_to(None))
        assert_that(if1.interface.shutdown, equal_to(True))
        assert_that(if1.interface.port_mode, equal_to(ACCESS))
        assert_that(if1.interface.access_vlan, equal_to(1999))
        assert_that(if1.interface.trunk_native_vlan, equal_to(None))
        assert_that(if1.interface.trunk_vlans, equal_to([]))
        assert_that(if1.members, equal_to([]))

    def test_get_bond_v2(self):
        self.requests_mock.should_receive("get").once().with_args(
            url=self.netman_url+'/switches/toto/bonds/3',
            headers=self.headers
        ).and_return(
            Reply(
                headers={'Netman-Version': '2'},
                content=open_fixture('get_switch_hostname_bond_v2.json').read(),
                status_code=200))

        if1 = self.switch.get_bond(3)

        assert_that(if1.number, equal_to(3))
        assert_that(if1.link_speed, equal_to('1g'))
        assert_that(if1.shutdown, equal_to(True))
        assert_that(if1.port_mode, equal_to(ACCESS))
        assert_that(if1.access_vlan, equal_to(1999))
        assert_that(if1.trunk_native_vlan, equal_to(None))
        assert_that(if1.trunk_vlans, equal_to([]))
        assert_that(if1.members, equal_to([]))

    def test_get_bonds_v1(self):
        self.requests_mock.should_receive("get").once().with_args(
            url=self.netman_url+'/switches/toto/bonds',
            headers=self.headers
        ).and_return(
            Reply(
                content=open_fixture('get_switch_hostname_bonds_v1.json').read(),
                status_code=200))

        if1, if2, if3 = self.switch.get_bonds()

        assert_that(if1.number, equal_to(3))
        assert_that(if1.link_speed, equal_to('1g'))
        assert_that(if1.interface.name, equal_to(None))
        assert_that(if1.interface.shutdown, equal_to(True))
        assert_that(if1.interface.port_mode, equal_to(ACCESS))
        assert_that(if1.interface.access_vlan, equal_to(1999))
        assert_that(if1.interface.trunk_native_vlan, equal_to(None))
        assert_that(if1.interface.trunk_vlans, equal_to([]))
        assert_that(if1.members, equal_to([]))

        assert_that(if2.number, equal_to(4))
        assert_that(if2.members, equal_to(["ge-0/0/1", "ge-1/0/1"]))
        assert_that(if2.interface.name, equal_to(None))
        assert_that(if2.interface.shutdown, equal_to(False))
        assert_that(if2.interface.port_mode, equal_to(TRUNK))
        assert_that(if2.interface.access_vlan, equal_to(None))
        assert_that(if2.interface.trunk_native_vlan, equal_to(2999))
        assert_that(if2.interface.trunk_vlans, equal_to([3000, 3001, 3002]))

        assert_that(if3.number, equal_to(6))
        assert_that(if3.link_speed, equal_to('10g'))
        assert_that(if3.interface.name, equal_to(None))
        assert_that(if3.interface.shutdown, equal_to(False))
        assert_that(if3.interface.port_mode, equal_to(DYNAMIC))
        assert_that(if3.interface.access_vlan, equal_to(1999))
        assert_that(if3.interface.trunk_native_vlan, equal_to(2999))
        assert_that(if3.interface.trunk_vlans, equal_to([3000, 3001, 3002]))
        assert_that(if3.members, equal_to([]))

    def test_get_bonds_v2(self):
        self.requests_mock.should_receive("get").once().with_args(
            url=self.netman_url+'/switches/toto/bonds',
            headers=self.headers
        ).and_return(
            Reply(
                headers={'Netman-Version': '2'},
                content=open_fixture('get_switch_hostname_bonds_v2.json').read(),
                status_code=200))

        if1, if2, if3 = self.switch.get_bonds()

        assert_that(if1.number, equal_to(3))
        assert_that(if1.link_speed, equal_to('1g'))
        assert_that(if1.shutdown, equal_to(True))
        assert_that(if1.port_mode, equal_to(ACCESS))
        assert_that(if1.access_vlan, equal_to(1999))
        assert_that(if1.trunk_native_vlan, equal_to(None))
        assert_that(if1.trunk_vlans, equal_to([]))
        assert_that(if1.members, equal_to([]))

        assert_that(if2.number, equal_to(4))
        assert_that(if2.members, equal_to(["ge-0/0/1", "ge-1/0/1"]))
        assert_that(if2.shutdown, equal_to(False))
        assert_that(if2.port_mode, equal_to(TRUNK))
        assert_that(if2.access_vlan, equal_to(None))
        assert_that(if2.trunk_native_vlan, equal_to(2999))
        assert_that(if2.trunk_vlans, equal_to([3000, 3001, 3002]))
        assert_that(if2.members, equal_to(['ge-0/0/1', 'ge-1/0/1']))

        assert_that(if3.number, equal_to(6))
        assert_that(if3.link_speed, equal_to('10g'))
        assert_that(if3.shutdown, equal_to(False))
        assert_that(if3.port_mode, equal_to(DYNAMIC))
        assert_that(if3.access_vlan, equal_to(1999))
        assert_that(if3.trunk_native_vlan, equal_to(2999))
        assert_that(if3.trunk_vlans, equal_to([3000, 3001, 3002]))
        assert_that(if3.members, equal_to([]))

    def test_add_vlan(self):
        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches/toto/vlans',
            headers=self.headers,
            data=JsonData(number=2000, name="deux-milles")
        ).and_return(
            Reply(
                content='',
                status_code=201))

        self.switch.add_vlan(2000, name="deux-milles")

    def test_add_vlan_without_a_name(self):
        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches/toto/vlans',
            headers=self.headers,
            data=JsonData(number=2000)
        ).and_return(
            Reply(
                content='',
                status_code=201))

        self.switch.add_vlan(2000)

    def test_add_vlan_already_exist(self):
        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches/toto/vlans',
            headers=self.headers,
            data=JsonData(number=2000, name="deux-milles")
        ).and_return(
            Reply(
                content=json.dumps({
                    "error": "Vlan 2000 already exists",
                    "error-module": VlanAlreadyExist.__module__,
                    "error-class": VlanAlreadyExist.__name__
                }),
                status_code=409))

        with self.assertRaises(VlanAlreadyExist) as expect:
            self.switch.add_vlan(2000, name="deux-milles")

        assert_that(str(expect.exception), equal_to("Vlan 2000 already exists"))

    def test_remove_vlan(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches/toto/vlans/2000',
            headers=self.headers,
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.remove_vlan(2000)

    def test_put_access_groups_in(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/vlans/2500/access-groups/in',
            headers=self.headers,
            data='spaceless-string'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_vlan_access_group(2500, IN, "spaceless-string")

    def test_put_access_groups_out(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/vlans/2500/access-groups/out',
            headers=self.headers,
            data='spaceless-string'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_vlan_access_group(2500, OUT, "spaceless-string")

    def test_remove_access_groups_in(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches/toto/vlans/2500/access-groups/in',
            headers=self.headers
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.unset_vlan_access_group(2500, IN)

    def test_remove_access_groups_out(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches/toto/vlans/2500/access-groups/out',
            headers=self.headers
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.unset_vlan_access_group(2500, OUT)

    def test_add_ip_to_vlan(self):
        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches/toto/vlans/2500/ips',
            headers=self.headers,
            data="1.2.3.4/25"
        ).and_return(
            Reply(
                content='',
                status_code=201))

        self.switch.add_ip_to_vlan(2500, ExactIpNetwork("1.2.3.4", 25))

    def test_remove_ip(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches/toto/vlans/2500/ips/1.2.3.4/25',
            headers=self.headers
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.remove_ip_from_vlan(2500, ExactIpNetwork("1.2.3.4", 25))

    def test_set_vlan_vrf(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/vlans/2500/vrf-forwarding',
            headers=self.headers,
            data="DEFAULT_LAN"
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_vlan_vrf(2500, "DEFAULT_LAN")

    def test_unset_vlan_vrf(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches/toto/vlans/2500/vrf-forwarding',
            headers=self.headers
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.unset_vlan_vrf(2500)

    def test_port_mode_access(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/port-mode',
            headers=self.headers,
            data='access'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_access_mode("ge-0/0/6")

    def test_port_mode_trunk(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/port-mode',
            headers=self.headers,
            data='trunk'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_trunk_mode("ge-0/0/6")

    def test_bond_port_mode_access(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/bonds/123/port-mode',
            headers=self.headers,
            data='access'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_bond_access_mode(123)

    def test_bond_port_mode_trunk(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/bonds/123/port-mode',
            headers=self.headers,
            data='trunk'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_bond_trunk_mode(123)

    def test_set_access_vlan(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/access-vlan',
            headers=self.headers,
            data='1000'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_access_vlan("ge-0/0/6", 1000)

    def test_reset_interface(self):
        self.requests_mock.should_receive("put").once().with_args(
                url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6',
                headers=self.headers,
                data=None
        ).and_return(
                Reply(
                        content='',
                        status_code=204))

        self.switch.reset_interface("ge-0/0/6")

    def test_reset_interface_with_unknown_interface_raises(self):
        self.requests_mock.should_receive("put").once().with_args(
                url=self.netman_url+'/switches/toto/interfaces/ne-0/0/66',
                headers=self.headers,
                data=None
        ).and_return(
                Reply(
                        content=json.dumps({
                            "error": "Interface ethernet ne-0/0/66 not found",
                            "error-module": UnknownInterface.__module__,
                            "error-class": UnknownInterface.__name__
                        }),
                        status_code=404))

        with self.assertRaises(UnknownInterface):
            self.switch.reset_interface('ne-0/0/66')

    def test_unset_interface_access_vlan(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/access-vlan',
            headers=self.headers
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.unset_interface_access_vlan("ge-0/0/6")

    def test_set_interface_native_vlan(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/trunk-native-vlan',
            headers=self.headers,
            data='1000'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_interface_native_vlan("ge-0/0/6", 1000)

    def test_unset_interface_native_vlan(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/trunk-native-vlan',
            headers=self.headers,
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.unset_interface_native_vlan("ge-0/0/6")

    def test_set_bond_native_vlan(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/bonds/123/trunk-native-vlan',
            headers=self.headers,
            data='1000'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_bond_native_vlan(123, 1000)

    def test_unset_bond_native_vlan(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches/toto/bonds/123/trunk-native-vlan',
            headers=self.headers,
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.unset_bond_native_vlan(123)

    def test_add_trunk_vlan(self):
        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/trunk-vlans',
            headers=self.headers,
            data='1000'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.add_trunk_vlan("ge-0/0/6", 1000)

    def test_remove_trunk_vlan(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/trunk-vlans/1000',
            headers=self.headers
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.remove_trunk_vlan("ge-0/0/6", 1000)

    def test_add_bond_trunk_vlan(self):
        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches/toto/bonds/123/trunk-vlans',
            headers=self.headers,
            data='1000'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.add_bond_trunk_vlan(123, 1000)

    def test_remove_bond_trunk_vlan(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches/toto/bonds/123/trunk-vlans/1000',
            headers=self.headers
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.remove_bond_trunk_vlan(123, 1000)

    def test_set_interface_description(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/description',
            headers=self.headers,
            data='Resistance is futile'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_interface_description("ge-0/0/6", "Resistance is futile")

    def test_unset_interface_description(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/description',
            headers=self.headers
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.unset_interface_description("ge-0/0/6")

    def test_set_bond_description(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/bonds/123/description',
            headers=self.headers,
            data='Resistance is futile'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_bond_description(123, "Resistance is futile")

    def test_unset_bond_description(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches/toto/bonds/123/description',
            headers=self.headers
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.unset_bond_description(123)

    def test_set_interface_mtu(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/mtu',
            headers=self.headers,
            data='5000'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_interface_mtu("ge-0/0/6", 5000)

    def test_unset_interface_mtu(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/mtu',
            headers=self.headers
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.unset_interface_mtu("ge-0/0/6")

    def test_set_bond_mtu(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/bonds/123/mtu',
            headers=self.headers,
            data='5000'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_bond_mtu(123, 5000)

    def test_unset_bond_mtu(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches/toto/bonds/123/mtu',
            headers=self.headers
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.unset_bond_mtu(123)

    def test_edit_interface_spanning_tree_succeeds(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/spanning-tree',
            headers=self.headers,
            data=json.dumps({"edge": True})
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.edit_interface_spanning_tree("ge-0/0/6", edge=True)

    def test_edit_interface_spanning_tree_optional_params(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/spanning-tree',
            headers=self.headers,
            data=json.dumps({})
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.edit_interface_spanning_tree("ge-0/0/6")

    def test_enable_interface(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/shutdown',
            headers=self.headers,
            data='false'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_interface_state("ge-0/0/6", ON)

    def test_disable_interface(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/shutdown',
            headers=self.headers,
            data='true'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_interface_state("ge-0/0/6", OFF)

    def test_unset_interface_state(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/shutdown',
            headers=self.headers
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.unset_interface_state("ge-0/0/6")

    def test_enable_interface_auto_negotiation(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/auto-negotiation',
            headers=self.headers,
            data='true'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_interface_auto_negotiation_state("ge-0/0/6", ON)

    def test_disable_interface_auto_negotiation(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/auto-negotiation',
            headers=self.headers,
            data='false'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_interface_auto_negotiation_state("ge-0/0/6", OFF)

    def test_unset_interface_auto_negotiation_state(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/auto-negotiation',
            headers=self.headers
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.unset_interface_auto_negotiation_state("ge-0/0/6")

    def test_add_bond(self):
        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches/toto/bonds',
            headers=self.headers,
            data=JsonData(number=6)
        ).and_return(
            Reply(
                content='',
                status_code=201))

        self.switch.add_bond(6)

    def test_remove(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches/toto/bonds/6',
            headers=self.headers
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.remove_bond(6)

    def test_add_interface_to_bond(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/bond-master',
            headers=self.headers,
            data='10'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.add_interface_to_bond('ge-0/0/6', 10)

    def test_remove_interface_from_bond(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/bond-master',
            headers=self.headers
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.remove_interface_from_bond('ge-0/0/6')

    def test_edit_bond_spanning_tree(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/bonds/5/spanning-tree',
            headers=self.headers,
            data=json.dumps({"edge": True})
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.edit_bond_spanning_tree(5, edge=True)

    def edit_bond_spanning_tree_optional_params(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/bonds/5/spanning-tree',
            headers=self.headers,
            data=json.dumps({})
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.edit_bond_spanning_tree(5)

    def test_change_bond_speed(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/bonds/10/link-speed',
            headers=self.headers,
            data='1g'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_bond_link_speed(10, '1g')

    def test_change_bond_speed_missing_bond(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/bonds/10/link-speed',
            headers=self.headers,
            data='1g'
        ).and_return(
            Reply(
                content=json.dumps({
                    "error": "Bond 10 not found",
                    "error-module": UnknownBond.__module__,
                    "error-class": UnknownBond.__name__
                }),
                status_code=404))

        with self.assertRaises(UnknownBond) as expect:
            self.switch.set_bond_link_speed(10, '1g')

        assert_that(str(expect.exception), equal_to("Bond 10 not found"))

    def test_change_bond_speed_wrong_value(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/bonds/10/link-speed',
            headers=self.headers,
            data='1z'
        ).and_return(
            Reply(
                content=json.dumps({
                    "error": "Malformed bond link speed",
                    "error-module": BadBondLinkSpeed.__module__,
                    "error-class": BadBondLinkSpeed.__name__
                }),
                status_code=400))

        with self.assertRaises(BadBondLinkSpeed) as expect:
            self.switch.set_bond_link_speed(10, '1z')

        assert_that(str(expect.exception), equal_to("Malformed bond link speed"))

    def test_change_bond_speed_switch_locked(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/bonds/10/link-speed',
            headers=self.headers,
            data='1z'
        ).and_return(
            Reply(
                content=json.dumps({
                    "error": "Switch is locked and can't be modified",
                    "error-module": LockedSwitch.__module__,
                    "error-class": LockedSwitch.__name__
                }),
                status_code=423))

        with self.assertRaises(LockedSwitch) as expect:
            self.switch.set_bond_link_speed(10, '1z')

        assert_that(str(expect.exception), equal_to("Switch is locked and can't be modified"))

    def test_add_vrrp_group(self):
        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches/toto/vlans/2000/vrrp-groups',
            headers=self.headers,
            data=JsonData(id=1,
                          priority=2,
                          ips=['1.2.3.4'],
                          hello_interval=5,
                          dead_interval=15,
                          track_id="101",
                          track_decrement=50)
        ).and_return(
            Reply(
                content='',
                status_code=201))

        self.switch.add_vrrp_group(2000, group_id=1, priority=2, ips=[IPAddress('1.2.3.4')], hello_interval=5,
                                   dead_interval=15, track_id='101', track_decrement=50)

    def test_remove_vrrp_group(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches/toto/vlans/2000/vrrp-groups/123',
            headers=self.headers
        ).and_return(
            Reply(
                content='',
                status_code=201))

        self.switch.remove_vrrp_group(2000, group_id=123)

    def test_add_dhcp_relay_server(self):
        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches/toto/vlans/2000/dhcp-relay-server',
            headers=self.headers,
            data='1.2.3.4'
        ).and_return(
            Reply(
                content='',
                status_code=201))

        self.switch.add_dhcp_relay_server(2000, '1.2.3.4')

    def test_remove_dhcp_relay_server(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches/toto/vlans/2000/dhcp-relay-server/1.2.3.4',
            headers=self.headers
        ).and_return(
            Reply(
                content='',
                status_code=201))

        self.switch.remove_dhcp_relay_server(2000, '1.2.3.4')

    def test_add_varp_ip(self):
        self.requests_mock.should_receive("post").once().with_args(
            url=self.netman_url+'/switches/toto/vlans/2000/varp-ips',
            headers=self.headers,
            data='1.2.3.4/29'
        ).and_return(
            Reply(
                content='',
                status_code=201))

        self.switch.add_vlan_varp_ip(2000, IPNetwork('1.2.3.4/29'))

    def test_remove_varp_ip(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url+'/switches/toto/vlans/2000/varp-ips/1.2.3.4/29',
            headers=self.headers
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.remove_vlan_varp_ip(2000, IPNetwork('1.2.3.4/29'))

    def test_set_interface_lldp_state(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/lldp',
            headers=self.headers,
            data='true'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_interface_lldp_state("ge-0/0/6", True)

    def test_disable_lldp(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/interfaces/ge-0/0/6/lldp',
            headers=self.headers,
            data='false'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_interface_lldp_state("ge-0/0/6", False)

    def test_set_vlan_icmp_redirects_state_False_should_send_false(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/vlans/2000/icmp-redirects',
            headers=self.headers,
            data='false'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_vlan_icmp_redirects_state(2000, False)

    def test_set_vlan_icmp_redirects_state_True_should_send_true(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url+'/switches/toto/vlans/2000/icmp-redirects',
            headers=self.headers,
            data='true'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_vlan_icmp_redirects_state(2000, True)

    def test_get_versions(self):
        data = {
            "v": "1.0",
            "units": {
                "1": {
                    "v": "1.0"
                }
            }
        }

        self.requests_mock.should_receive("get").once().with_args(
            url=self.netman_url+'/switches/toto/versions',
            headers=self.headers
        ).and_return(
            Reply(
                content=json.dumps(data),
                status_code=204))

        result = self.switch.get_versions()
        assert_that(result, is_(data))

    def test_unformatted_exceptions_are_handled(self):
        self.requests_mock.should_receive("put").once().and_return(Reply(
            content='Oops an unexpected excepton occured',
            status_code=500
        ))

        with self.assertRaises(Exception) as expect:
            self.switch.set_bond_link_speed(10, '1z')

        assert_that(str(expect.exception), equal_to("500: Oops an unexpected excepton occured"))

    def test_native_exceptions_are_handled(self):
        self.requests_mock.should_receive("put").once().and_return(Reply(
            content=json.dumps({
                "error": "Oops an unexpected excepton occured",
                "error-class": "Exception"
            }),
            status_code=500
        ))

        with self.assertRaises(Exception) as expect:
            self.switch.set_bond_link_speed(10, '1z')

        assert_that(str(expect.exception), equal_to("Oops an unexpected excepton occured"))

    def test_exceptions_missing_error_classes_work(self):
        self.requests_mock.should_receive("put").once().and_return(Reply(
            content=json.dumps({
                "error": "Oops an unexpected excepton occured"
            }),
            status_code=500
        ))

        with self.assertRaises(Exception) as expect:
            self.switch.set_bond_link_speed(10, '1z')

        assert_that(str(expect.exception), equal_to("Oops an unexpected excepton occured"))

    def test_exceptions_bad_init_works(self):
        self.requests_mock.should_receive("put").once().and_return(Reply(
            content=json.dumps({
                "error": "Switch is locked and can't be modified",
                "error-module": RPCError.__module__,
                "error-class": RPCError.__name__
            }),
            status_code=400
        ))

        with self.assertRaises(NetmanException) as expect:
            self.switch.set_bond_link_speed(10, '1z')

        assert_that(
            str(expect.exception),
            equal_to("ncclient.operations.rpc.RPCError: Switch is locked and can't be modified"))

    def test_set_vlan_load_interval(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url + "/switches/toto/vlans/123/load-interval",
            headers=self.headers,
            data="30"
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_vlan_load_interval(123, 30)

    def test_unset_vlan_load_interval(self):
        self.requests_mock.should_receive("delete").once().with_args(
            url=self.netman_url + "/switches/toto/vlans/123/load-interval",
            headers=self.headers
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.unset_vlan_load_interval(123)

    def test_set_vlan_mpls_ip_state_true(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url + "/switches/toto/vlans/123/mpls-ip",
            headers=self.headers,
            data='true'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_vlan_mpls_ip_state(123, True)

    def test_set_vlan_mpls_ip_state_false(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url + "/switches/toto/vlans/123/mpls-ip",
            headers=self.headers,
            data='false'
        ).and_return(
            Reply(
                content='',
                status_code=204))

        self.switch.set_vlan_mpls_ip_state(123, False)

    def test_set_vlan_mpls_ip_state_invalid(self):
        self.requests_mock.should_receive("put").once().with_args(
            url=self.netman_url + "/switches/toto/vlans/123/mpls-ip",
            headers=self.headers,
            data='30'
        ).and_return(
            Reply(
                content=json.dumps({"error": "",
                                    "error-module": BadMplsIpState.__module__,
                                    "error-class": BadMplsIpState.__name__}),
                status_code=400))

        with self.assertRaises(BadMplsIpState):
            self.switch.set_vlan_mpls_ip_state(123, 30)


class Reply:
    def __init__(self, status_code, content, headers=None):
        self.status_code = status_code
        self.content = content
        self.headers = headers or {}

    def json(self):
        return json.loads(self.content)


class JsonData:
    def __init__(self, **data):
        self.data = data

    def __eq__(self, other):
        try:
            return json.loads(other) == self.data
        except ValueError:
            return False



# Copyright 2017 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from flexmock import flexmock, flexmock_teardown
from hamcrest import assert_that, equal_to

from netman.adapters.switches.cisco import Cisco
from netman.adapters.switches.util import SubShell
from netman.core.objects.exceptions import UnknownVlan
from netman.core.objects.interface_states import OFF, ON
from netman.core.objects.switch_descriptor import SwitchDescriptor


class CiscoArpRoutingTest(unittest.TestCase):
    def setUp(self):
        self.switch = Cisco(SwitchDescriptor(model='cisco', hostname="my.hostname"))
        SubShell.debug = True
        self.mocked_ssh_client = flexmock()
        self.switch.ssh = self.mocked_ssh_client

    def tearDown(self):
        flexmock_teardown()

    def test_set_vlan_arp_routing_state_disable(self):
        self.mocked_ssh_client.should_receive("do").with_args("show running-config interface vlan 1234")\
            .once()\
            .ordered()\
            .and_return(["Building configuration...",
                         "Current configuration : 41 bytes",
                         "!",
                         "interface Vlan1234",
                         " no ip address",
                         "end"])

        self.mocked_ssh_client.should_receive("do").with_args("configure terminal").once().ordered().and_return([
            "Enter configuration commands, one per line.  End with CNTL/Z."
        ])
        self.mocked_ssh_client.should_receive("do").with_args("interface vlan 1234").and_return([]).once().ordered()
        self.mocked_ssh_client.should_receive("do").with_args("no shutdown").and_return([]).once().ordered()
        self.mocked_ssh_client.should_receive("do").with_args("no ip proxy-arp").and_return([]).once().ordered()
        self.mocked_ssh_client.should_receive("do").with_args("exit").and_return([]).twice().ordered().ordered()

        self.switch.set_vlan_arp_routing_state(1234, OFF)

    def test_set_vlan_arp_routing_state_enable(self):
        self.mocked_ssh_client.should_receive("do").with_args("show running-config interface vlan 1234").once().ordered()\
            .and_return([
                "Building configuration...",
                "Current configuration : 41 bytes",
                "!",
                "interface Vlan1234",
                " no ip address",
                "end"])

        self.mocked_ssh_client.should_receive("do").with_args("configure terminal").once().ordered().and_return([
            "Enter configuration commands, one per line.  End with CNTL/Z."
        ])
        self.mocked_ssh_client.should_receive("do").with_args("interface vlan 1234").and_return([]).once().ordered()
        self.mocked_ssh_client.should_receive("do").with_args("no shutdown").and_return([]).once().ordered()
        self.mocked_ssh_client.should_receive("do").with_args("ip proxy-arp").and_return([]).once().ordered()
        self.mocked_ssh_client.should_receive("do").with_args("exit").and_return([]).twice().ordered().ordered()

        self.switch.set_vlan_arp_routing_state(1234, ON)

    def test_set_vlan_arp_routing_state_without_interface_creates_it(self):
        self.mocked_ssh_client.should_receive("do").with_args(
            "show running-config interface vlan 1234").once().ordered().and_return([
                "                                  ^",
                "% Invalid input detected at '^' marker."])

        self.mocked_ssh_client.should_receive("do").with_args(
            "show running-config vlan 1234 | begin vlan")\
            .once().ordered().and_return(["vlan 1234", "end"])

        self.mocked_ssh_client.should_receive("do").with_args("configure terminal").once().ordered().and_return([
            "Enter configuration commands, one per line.  End with CNTL/Z."
        ])
        self.mocked_ssh_client.should_receive("do").with_args("interface vlan 1234").and_return([]).once().ordered()
        self.mocked_ssh_client.should_receive("do").with_args("no shutdown").and_return([]).once().ordered()
        self.mocked_ssh_client.should_receive("do").with_args("no ip proxy-arp").and_return([]).once().ordered()
        self.mocked_ssh_client.should_receive("do").with_args("exit").and_return([]).twice().ordered().ordered()
        self.switch.set_vlan_arp_routing_state(1234, OFF)

    def test_set_vlan_arp_routing_state_unknown_vlan(self):
        self.mocked_ssh_client.should_receive("do").with_args(
            "show running-config interface vlan 1234")\
            .once().ordered().and_return(["                                  ^",
                                          "% Invalid input detected at '^' marker."])

        self.mocked_ssh_client.should_receive("do").with_args(
            "show running-config vlan 1234 | begin vlan").once().ordered().and_return([])

        with self.assertRaises(UnknownVlan) as expect:
            self.switch.set_vlan_arp_routing_state(1234, OFF)

        assert_that(str(expect.exception), equal_to("Vlan 1234 not found"))
